####################################################################################################
# neuropythy/graphics/core.py
# Core implementation of the neuropythy graphics library for plotting cortical surfaces

import numpy               as np
import numpy.linalg        as npla
import scipy               as sp
import scipy.sparse        as sps
import scipy.spatial       as spspace
import neuropythy.geometry as geo
import neuropythy.mri      as mri
import neuropythy.io       as nyio
import os, sys, six, itertools, atexit, shutil, tempfile, warnings, pimms

from ..util            import (times, zdivide, plus, minus, to_hemi_str, nanlog)
from ..vision          import (visual_area_names, visual_area_numbers)

# 2D Graphics ######################################################################################

# Below are various graphics-related functions; some are wrapped in a try block in case matplotlib
# is not installed.
def _matplotlib_load_error(*args, **kwargs):
    raise RuntimeError('load failure: the requested object could not be loaded, probably because ' +
                       'you do not have matplotlib installed correctly')
cmap_curvature = _matplotlib_load_error
cmap_polar_angle_sym = _matplotlib_load_error
cmap_polar_angle_lh = _matplotlib_load_error
cmap_polar_angle_rh = _matplotlib_load_error
cmap_polar_angle = _matplotlib_load_error
cmap_theta_sym = _matplotlib_load_error
cmap_theta_lh = _matplotlib_load_error
cmap_theta_rh = _matplotlib_load_error
cmap_theta = _matplotlib_load_error
cmap_eccentricity = _matplotlib_load_error
cmap_log_eccentricity = _matplotlib_load_error
cmap_radius = _matplotlib_load_error
cmap_log_radius = _matplotlib_load_error
cmap_cmag = _matplotlib_load_error
cmap_log_cmag = _matplotlib_load_error
cmap_eccenflat = _matplotlib_load_error
label_cmap = _matplotlib_load_error

try:
    import matplotlib, matplotlib.pyplot, matplotlib.tri, matplotlib.colors
    # we use this below
    blend_cmap = matplotlib.colors.LinearSegmentedColormap.from_list
    
    cmap_curvature = matplotlib.colors.LinearSegmentedColormap(
        'curv',
        {name: ((0.0, 0.0, 0.5), (0.5, 0.5, 0.2), (1.0, 0.2, 0.0))
         for name in ['red', 'green', 'blue']})
    cmap_curvature.__doc__ = '''
    cmap_curvature is a colormap for plotting the curvature of a vertex; note that this colormap
    assumes FreeSurfer's standard way of storing curvature where negative values indicate gyri and
    positive values indicate sulci.
    Values passed to cmap_curvature should be scaled such that (-1,1) -> (0,1).
    '''
    
    cmap_polar_angle_sym = blend_cmap(
        'polar_angle_sym',
        [(0.5,0,0), (1,1,0), (0,0.5,0), (0,1,1), (0,0,0.5), (0,1,1), (0,0.5,0), (1,1,0), (0.5,0,0)])
    cmap_polar_angle_sym.__doc__ = '''
    cmap_polar_angle_sym is a colormap for plotting the pRF polar angle of a vertex.
    Values passed to cmap_polar_angle_sym should be scaled such that (-180,180 deg) -> (0,1).

    cmap_polar_angle_sym is a circular colormap that is left-right symmetric with green representing
    the horizontal meridian, blue representing the upper vertical meridian, and red representing the
    lower vertical meridian.
    '''
    cmap_polar_angle_lh = blend_cmap(
        'polar_angle_lh',
        [(0.5,0,0), (0.75,0,0.5), (1,0,1), (0.5,0,0.75), (0,0,0.5), (0,1,1), (0,0.5,0), (1,1,0),
         (0.5,0,0)])
    cmap_polar_angle_lh.__doc__ = '''
    cmap_polar_angle_lh is a colormap for plotting the pRF polar angle of a vertex.
    Values passed to cmap_polar_angle_lh should be scaled such that (-180,180 deg) -> (0,1).

    cmap_polar_angle_lh is a circular colormap that emphasizes colors in the right visual field; the
    left visual field appears mostly magenta.
    '''
    cmap_polar_angle_rh = blend_cmap(
        'polar_angle_rh',
        [(0.5,0,0), (1,1,0), (0,0.5,0), (0,1,1), (0,0,0.5), (0.5,0,0.75), (1,0,1), (0.75,0,0.5),
         (0.5,0,0)])
    cmap_polar_angle_rh.__doc__ = '''
    cmap_polar_angle_rh is a colormap for plotting the pRF polar angle of a vertex.
    Values passed to cmap_polar_angle_rh should be scaled such that (-180,180 deg) -> (0,1).

    cmap_polar_angle_rh is a circular colormap that emphasizes colors in the left visual field; the
    right visual field appears mostly magenta.
    '''
    cmap_polar_angle = blend_cmap(
        'polar_angle',
         [(0.5,0,0), (1,0,1), (0,0,0.5), (0,1,1), (0,0.5,0), (1,1,0), (0.5,0,0)])
    cmap_polar_angle.__doc__ = '''
    cmap_polar_angle is a colormap for plotting the pRF polar angle of a vertex.
    Values passed to cmap_polar_angle should be scaled such that (-180,180 deg) -> (0,1).

    cmap_polar_angle is a 6-pronged circular colormap; note that it does not have dark or bright
    values at the horizontal meridia as cmap_polar_angle_sym, cmap_polar_angle_lh, and
    cmap_polar_angle_rh do.
    '''
    cmap_theta_sym = blend_cmap(
        'theta_sym',
        [(0.5,0,0), (1,1,0), (0,0.5,0), (0,1,1), (0,0,0.5), (0,1,1), (0,0.5,0), (1,1,0), (0.5,0,0)])
    cmap_theta_sym.__doc__ = '''
    cmap_theta_sym is a colormap for plotting the pRF theta of a vertex.
    Values passed to cmap_theta_sym should be scaled such that (-pi,pi rad) -> (0,1). Note that 0 is
    the right right horizontal meridian and positive is the counter-clockwise direction.

    cmap_theta_sym is a circular colormap that is left-right symmetric with green representing
    the horizontal meridian, blue representing the upper vertical meridian, and red representing the
    lower vertical meridian.
    '''
    cmap_theta_lh = blend_cmap(
        'theta_lh',
        [(0.5,0,0), (1,1,0), (0,0.5,0), (0,1,1), (0,0,0.5),
         (0.5,0,0.75), (1,0,1), (0.75,0,0.5), (0.5,0,0)])
    cmap_theta_lh.__doc__ = '''
    cmap_theta_lh is a colormap for plotting the pRF theta of a vertex.
    Values passed to cmap_theta_lh should be scaled such that (-pi,pi rad) -> (0,1). Note that 0 is
    the right right horizontal meridian and positive is the counter-clockwise direction.

    cmap_theta_lh is a circular colormap that emphasizes colors in the right visual field; the
    left visual field appears mostly magenta.
    '''
    cmap_theta_rh = blend_cmap(
        'theta_rh',
        [(0.5,0,0), (0.75,0,0.5), (0.5,0,0.75), (1,0,1),
         (0,0,0.5), (0,1,1), (0,0.5,0), (1,1,0), (0.5,0,0)])
    cmap_theta_rh.__doc__ = '''
    cmap_theta_rh is a colormap for plotting the pRF theta of a vertex.
    Values passed to cmap_theta_rh should be scaled such that (-pi,pi rad) -> (0,1). Note that 0 is
    the right right horizontal meridian and positive is the counter-clockwise direction.

    cmap_theta_rh is a circular colormap that emphasizes colors in the left visual field; the
    right visual field appears mostly magenta.
    '''
    cmap_theta = blend_cmap(
        'theta',
        [(0,       cmap_polar_angle(0.25)),
         ( 1.0/12, (  1,  0,  1)),
         ( 3.0/12, (0.5,  0,  0)),
         ( 5.0/12, (  1,  1,  0)),
         ( 7.0/12, (  0,0.5,  0)),
         ( 9.0/12, (  0,  1,  1)),
         (11.0/12, (  0,  0,0.5)),
         (1,       cmap_polar_angle(0.25))])
    cmap_theta.__doc__ = '''
    cmap_theta is a colormap for plotting the pRF theta of a vertex.
    Values passed to cmap_theta should be scaled such that (-pi,pi rad) -> (0,1). Note that 0 is the
    right right horizontal meridian and positive is the counter-clockwise direction.

    cmap_theta is a 6-pronged circular colormap; note that it does not have dark or bright
    values at the horizontal meridia as cmap_theta_sym, cmap_theta_lh, and
    cmap_theta_rh do.
    '''
    cmap_eccentricity = blend_cmap(
        'eccentricity',
        [(0,       (  0,  0,  0)),
         (1.25/90, (  0,  0,0.5)),
         (2.5/90,  (  1,  0,  1)),
         (5.0/90,  (0.5,  0,  0)),
         (10.0/90, (  1,  1,  0)),
         (20.0/90, (  0,0.5,  0)),
         (40.0/90, (  0,  1,  1)),
         (1,       (  1,  1,  1))])
    cmap_eccentricity.__doc__ = '''
    cmap_eccentricity is a colormap for plotting the pRF eccentricity of a vertex.
    Values passed to cmap_eccentricity should be scaled such that (0,90 deg) -> (0,1).
    '''
    cmap_log_eccentricity = blend_cmap(
        'log_eccentricity',
        [(0,     (  0,  0,  0)),
         (1.0/7, (  0,  0,0.5)),
         (2.0/7, (  1,  0,  1)),
         (3.0/7, (0.5,  0,  0)),
         (4.0/7, (  1,  1,  0)),
         (5.0/7, (  0,0.5,  0)),
         (6.0/7, (  0,  1,  1)),
         (1,     (  1,  1,  1))])
    cmap_log_eccentricity.__doc__ = '''
    cmap_log_eccentricity is a colormap for plotting the log of eccentricity.
    Values passed to cmap_log_cmag should be scaled however desired, but note that the colormap
    itself runs linearly from 0 to 1, so eccentricity data should be log-transformed
    then scaled before being passed.
    '''
    cmap_radius = blend_cmap(
        'radius',
        [(0,       (  0,  0,  0)),
         (1.25/60, (  0,  0,0.5)),
         (1.25/30, (  1,  0,  1)),
         ( 2.5/30, (0.5,  0,  0)),
         ( 5.0/30, (  1,  1,  0)),
         (10.0/30, (  0,0.5,  0)),
         (20.0/30, (  0,  1,  1)),
         (1,       (  1,  1,  1))])
    cmap_radius.__doc__ = '''
    cmap_radius is a colormap for plotting the pRF radius (sigma) of a vertex.
    Values passed to cmap_radius should be scaled such that (0,30 deg) -> (0,1).
    '''
    cmap_log_radius = blend_cmap(
        'log_radius',
        [(0,     (  0,  0,  0)),
         (1.0/7, (  0,  0,0.5)),
         (2.0/7, (  1,  0,  1)),
         (3.0/7, (0.5,  0,  0)),
         (4.0/7, (  1,  1,  0)),
         (5.0/7, (  0,0.5,  0)),
         (6.0/7, (  0,  1,  1)),
         (1,     (  1,  1,  1))])
    cmap_log_radius.__doc__ = '''
    cmap_log_radius is a colormap for plotting the log of radius.
    Values passed to cmap_log_cmag should be scaled however desired, but note that the colormap
    itself runs linearly from 0 to 1, so radius data should be log-transformed then scaled before
    being passed.
    '''
    cmap_cmag = blend_cmap(
        'cmag',
        [(0,         (  0,  0,  0)),
         (0.25/256,  (  0,  0,  0)),
         (1.23/256,  (  0,  0,0.5)),
         (2.97/256,  (  1,  0,  1)),
         (7.25/256,  (0.5,  0,  0)),
         (17.69/256, (  1,  1,  0)),
         (43.13/256, (  0,0.5,  0)),
         (105.0/256, (  0,  1,  1)),
         (1,         (  1,  1,  1))])
    cmap_cmag.__doc__ = '''
    cmap_cmag is a colormap for plotting cortical magnification.
    It is generally advised to use the cmap_log_cmag colormap, as it will give better results for
    the logarithmically-distributed corotical magnification colormap.
    '''
    cmap_log_cmag = blend_cmap(
        'log_cmag',
        [(0,     (  0,  0,  0)),
         (1.0/7, (  0,  0,0.5)),
         (2.0/7, (  1,  0,  1)),
         (3.0/7, (0.5,  0,  0)),
         (4.0/7, (  1,  1,  0)),
         (5.0/7, (  0,0.5,  0)),
         (6.0/7, (  0,  1,  1)),
         (1,     (  1,  1,  1))])
    cmap_log_cmag.__doc__ = '''
    cmap_log_cmag is a colormap for plotting the log of cortical magnification.
    Values passed to cmap_log_cmag should be scaled however desired, but note that the colormap
    itself runs linearly from 0 to 1, so cortical magnification data should be log-transformed
    then scaled before being passed.
    '''
    cmap_eccenflat = blend_cmap(
        'eccenflat',
        [(0,     (  0,  0,  0)),
         (1.0/7, (  0,  0,0.5)),
         (2.0/7, (  1,  0,  1)),
         (3.0/7, (0.5,  0,  0)),
         (4.0/7, (  1,  1,  0)),
         (5.0/7, (  0,0.5,  0)),
         (6.0/7, (  0,  1,  1)),
         (1,     (  1,  1,  1))])
    cmap_eccenflat.__doc__ = '''
    cmap_eccenflat is a colormap for plotting the log of eccentricity; in fact, it is identical
    to the cmap_log_eccentricity colorscale, but when used with neuropythy it is assumed to be
    linear instead of logarithmic.
    '''
    # A few other handy colormaps:
    cmap_temperature_dark = blend_cmap('temperature_dark',
                                       [(0,1,1), (0,0,1), (0,0,0), (1,0,0), (1,1,0)])
    cmap_temperature      = blend_cmap('temperature',
                                       [(0,0,1), (0,1,1), (1,1,1), (1,1,0), (1,0,0)])
    cmap_lightsaber_dark  = blend_cmap('lightsaber_dark',
                                       [(0,1,1), (0,1,0), (0,0,0), (1,0,0), (1,1,0)])
    cmap_lightsaber       = blend_cmap('lightsaber',
                                       [(0,1,0), (0,1,1), (1,1,1), (1,1,0), (1,0,0)])
    cmap_electricity_dark = blend_cmap('electricity_dark',
                                       [(0,1,1), (0,0,1), (0,0,0), (0,1,0), (1,1,0)])
    cmap_electricity      = blend_cmap('electricity',
                                       [(0,0,1), (0,1,1), (1,1,1), (1,1,0), (0,1,0)])
    cmap_reddish          = blend_cmap('reddish',
                                       [(1,1,1), (1,1,0), (1,0,0), (0.5, 0, 0.25)])
    cmap_reddish_dark     = blend_cmap('reddish_dark',
                                       [(0,0,0), (1,0,0), (1,1,0), (1, 1, 0.25)])
    cmap_bluish           = blend_cmap('bluish',
                                       [(1,1,1), (0,1,1), (0,0,1), (0.25, 0, 0.5)])
    cmap_bluish_dark      = blend_cmap('bluish_dark',
                                       [(0,0,0), (0,0,1), (0,1,1), (0.25, 1, 1)])
    cmap_greenish         = blend_cmap('greenish',
                                       [(1,1,1), (0,1,1), (0,1,0), (0.25, 0.5, 0)])
    cmap_greenish_dark    = blend_cmap('greenish_dark',
                                       [(0,0,0), (0,1,0), (1,1,0), (1, 1, 0.5)])

    colormaps = {
        'curvature':        (cmap_curvature,        (-1,1)),
        'polar_angle':      (cmap_polar_angle,      (-180,180), 'deg'),
        'polar_angle_sym':  (cmap_polar_angle_sym,  (-180,180), 'deg'),
        'polar_angle_lh':   (cmap_polar_angle_lh,   (-180,180), 'deg'),
        'polar_angle_rh':   (cmap_polar_angle_rh,   (-180,180), 'deg'),
        'theta':            (cmap_theta,            (-np.pi,np.pi), 'rad'),
        'theta_sym':        (cmap_theta_sym,        (-np.pi,np.pi), 'rad'),
        'theta_lh':         (cmap_theta_lh,         (-np.pi,np.pi), 'rad'),
        'theta_rh':         (cmap_theta_rh,         (-np.pi,np.pi), 'rad'),
        'eccentricity':     (cmap_eccentricity,     (0,90), 'deg'),
        'log_eccentricity': (cmap_log_eccentricity, (np.log(0.75), np.log(90.75)), 'deg'),
        'radius':           (cmap_radius,           (0, 30), 'deg'), 
        'log_radius':       (cmap_log_radius,       (np.log(0.25), np.log(30.25)), 'deg'),
        'cmag2':            (cmap_cmag,             (0.25, 512.25), 'mm**2/deg**2'),
        'log_cmag2':        (cmap_log_cmag,         (np.log(0.25), np.log(512.25)), 'mm**2/deg**2'),
        'cmag':             (cmap_cmag,             (0.5, 256.5), 'mm/deg'),
        'log_cmag':         (cmap_log_cmag,         (np.log(0.5), np.log(256.5)), 'mm/deg'),
        'eccenflat':        (cmap_eccenflat,        (0, 1)),
        # the handy but non-neuroscience-based ones:
        'temperature':      (cmap_temperature,      (-1,1)),
        'temperature_dark': (cmap_temperature_dark, (-1,1)),
        'lightsaber':       (cmap_lightsaber,       (-1,1)),
        'lightsaber_dark':  (cmap_lightsaber_dark,  (-1,1)),
        'electricity':      (cmap_electricity,      (-1,1)),
        'electricity_dark': (cmap_electricity_dark, (-1,1)),
        'reddish':          (cmap_reddish,          (0,1)),
        'reddish_dark':     (cmap_reddish_dark,     (0,1)),
        'greenish':         (cmap_greenish,         (0,1)),
        'greenish_dark':    (cmap_greenish_dark,    (0,1)),
        'bluish':           (cmap_bluish,           (0,1)),
        'bluish_dark':      (cmap_bluish_dark,      (0,1))}
    for (k,cmdat) in six.iteritems(colormaps): matplotlib.cm.register_cmap(k, cmdat[0])

    def _diff_order(n):
        u0 = np.arange(n)
        d = int(np.ceil(np.sqrt(n)))
        mtx = np.reshape(np.pad(u0, [(0,d*d-n)], 'constant', constant_values=[(0,-1)]), (d,d))
        h = int((d+1)/2)
        u = np.vstack([mtx[::2], mtx[1::2]]).T.flatten()
        return u[u >= 0]
    def label_cmap(colors, cmap=None, name=Ellipsis):
        '''
    label_cmap(n) yields a colormap with n color-steps that should be optimized such that each
      colors is relatively different from its neighbors on the color spectrum. This is generally
      well-suited for discrete catgory/label colormaps.

    Note that this function uses a heuristic and is not guaranteed to be optimal in any way for any
    value of n--but it generally works well enough for most common purposes.
    
    The following optional arguments may be given:
      * cmap (default: None) specifies a colormap to use as a base. If this is None, then a varianct
        of 'hsv' is used.
      * name (default: None) specifies a name (string) that will be used for the colormap name; if
        Ellipsis is given, then ('label%d' % colors) is used.
    '''
        if not pimms.is_int(colors):
            (lbls,ris) = np.unique(colors, return_inverse=True)
            if lbls[0] == 0:
                lbls = lbls[1:]
                cm   = label_cmap(len(lbls), cmap=cmap)
                mx   = np.max(lbls)
                clrs = cm(np.linspace(0, 1, len(lbls)))
                return blend_cmap(name, [(0, [0,0,0,0])] + list(zip(lbls/mx, clrs)))
        if name is Ellipsis: name = 'label%d' % colors
        if pimms.is_str(cmap): cmap = getattr(mpl.cm, cmap)
        # get a diff-ordering
        u = _diff_order(colors)
        cm = matplotlib.cm.hsv if cmap is None else cmap
        clrs = cm(u / float(colors))
        if cmap is None and len(clrs) > 9:
            # we use the hsv-like map: equivalent to using hsv then modifying it with a saturation
            # and value label-cmap; this shouldn't go too low to prevent colors getting washed out
            d = int(np.ceil(np.sqrt(colors)))
            uu = _diff_order(d) / float(d-1)
            uu = (uu + 1) / 2
            ii = np.where(uu == 1)[0][0]
            uu = np.roll(uu, -ii, 0)
            uu = np.asarray([[x]*d for x in uu]).flatten()[:colors]
            # make sure the highest value comes first
            clrs[:,:3] *= np.reshape(uu, (-1,1))
        cm = blend_cmap(name, list(zip(np.linspace(0,1,colors), clrs)))
        return cm
except Exception: pass

def scale_for_cmap(cmap, x, vmin=Ellipsis, vmax=Ellipsis, unit=Ellipsis):
    '''
    scale_for_cmap(cmap, x) yields the values in x rescaled to be appropriate for the given
      colormap cmap. The cmap must be the name of a colormap or a colormap object.

    For a given cmap argument, if the object is a colormap itself, it is treated as cmap.name.
    If the cmap names a colormap known to neuropythy, neuropythy will rescale the values in x
    according to a heuristic.
    '''
    import matplotlib as mpl
    if isinstance(cmap, mpl.colors.Colormap): cmap = cmap.name
    (name, cm) = (None, None)
    if cmap not in colormaps:
        for (k,v) in six.iteritems(colormaps):
            if cmap in k:
                (name, cm) = (k, v)
                break
    else: (name, cm) = (cmap, colormaps[cmap])
    if cm is not None:
        cm = cm if len(cm) == 3 else (cm + (None,))
        (cm, (mn,mx), uu) = cm
        if vmin is Ellipsis: vmin = mn
        if vmax is Ellipsis: vmax = mx
        if unit is Ellipsis: unit = uu
    if vmin is Ellipsis: vmin = None
    if vmax is Ellipsis: vmax = None
    if unit is Ellipsis: unit = None
    x = pimms.mag(x) if unit is None else pimms.mag(x, unit)
    if name is not None and name.startswith('log_'):
        emn = np.exp(vmin)
        x = np.log(x + emn)
    vmin = np.nanmin(x) if vmin is None else vmin
    vmax = np.nanmax(x) if vmax is None else vmax
    return zdivide(x - vmin, vmax - vmin, null=np.nan)
def visual_field_legend(cmap, on=Ellipsis, max_eccentricity=12, transform=Ellipsis, pixels=288,
                        background=None, boundary_pixels=0):
    '''
    visual_field_map('polar_angle') yields an image array of a legend of the polar angle colormap,
      plotted in the visual-field.
    visual_field_map('eccentricity') yields an image array of a legend of the eccentricity colormap,
      plotted in the visual-field.
    visual_field_map(colormap, property) is used to color an arbitrary colormap on the given
      property. The prorperty may be 'polar_angle', 'eccentricity', 'theta', 'rho', 'x', or 'y'. For
      explanations of these properties, see below.

    Note: the call visual_field_legend('log_eccentricity') will plot an eccentricity legend using
    the log_eccentricity colormap; this is not a plot of the log-eccentricity, but rather is usually
    just a smoother/nicer version of the 'eccentricity' plot. This automatic scaling is performed
    only if the second parameter is not provided or is given as Ellipsis.

    The following optional arguments are accepted:
      * max_eccentricity (default: 12) specifies the maximum eccentricity of the plot; for polar
        angle plots this is not important, but for eccentricity, this will limit the range of the
        plot.
      * transform (default: Ellipsis) specifies a transform function f(x) that is applied to the
        given property before being passed to the colormap. Note that this function should accept
        and return a vector of values between 0 and 1 (for use by the colormap). A value of None
        indicates that no transformation should be used. The default value of Ellipsis indicates
        that the transformation should be detected automatically based on the plotted property; see
        below for default scaling.
      * pixels (default: 288) specifies the number of pixels in the width/height of the image.
      * background (default: None) specifies the background on which to plot the image. The default
        value of None is equivalent to (1,1,1,0) or a transparent background. Anything that, when
        passed to to_rgba() yields a single color-vector, can be passed to background.
      * boundary_pixels (default: 0) specifies the number of pixels around the edge of the image to
        reserve for the boundary. This may be specified as (left, right, bottom, top) or as
        (sides, top-bottom). In all cases, the boundary_pixels value does not change the width of
        the image but rather makes the plot smaller.
    '''
    # Start by parsing arguments:
    if max_eccentricity is None: max_eccentricity = 90
    if pimms.is_str(cmap):
        cmap = cmap.replace('-', '_').replace(' ', '_').lower()
        if on is Ellipsis:
            if   'polar_angle' in cmap: on = 'polar_angle'
            elif 'theta' in cmap: on = 'theta'
            elif 'log_eccentricity' in cmap:
                on = 'eccentricity'
                (mn,mx) = np.log([0.75, 0.75 + max_eccentricity])
                if transform is Ellipsis:
                    transform = lambda x: (np.log(x + 0.75) - mn) / (mx - mn)
            elif 'eccentricity' in cmap: on = 'eccentricity'
        if cmap in colormaps:
            (cm, (mn,mx)) = colormaps[cmap][:2]
            if transform is Ellipsis:
                if cmap.startswith('log'):
                    emn = np.exp(mn)
                    transform = lambda x: (np.log(x+emn) - mn) / (mx - mn)
                else:
                    transform = lambda x: (x - mn) / (mx - mn)
            cmap = cm
        else:
            cmap = getattr(matplotlib.cm, cmap)
    if transform is Ellipsis: transform = None
    if background is None: background = (1,1,1,0)
    else: background = to_rgba(background)
    # go ahead and make the image and the image-portion we are using
    whole_im = np.zeros((pixels, pixels, 4))
    whole_im[:,:,:] = np.reshape(background, (1,1,4))
    if boundary_pixels is None or boundary_pixels <= 0: im = whole_im
    else: im = whole_im[boundary_pixels:-boundary_pixels, boundary_pixels:-boundary_pixels, :]
    # note that this is visual field pixels:
    mid = im.shape[0] / 2
    (x_im, y_im) = np.meshgrid(np.arange(im.shape[0]) - mid, mid - np.arange(im.shape[1]))
    r_im = np.sqrt(x_im**2 + y_im**2)
    ii = np.where(r_im <= mid)
    x = x_im[ii] / mid * max_eccentricity
    y = y_im[ii] / mid * max_eccentricity
    # what property are we plotting on?
    if   on is Ellipsis: raise ValueError('could not deduce property on which to operate')
    elif not pimms.is_str(on): raise ValueError('plot-property must be a string')
    on = on.replace('-', '_').replace(' ', '_').lower()
    if   on == 'polar_angle':  p = np.mod(90 - 180/np.pi*np.arctan2(y, x) + 180, 360) - 180
    elif on == 'theta':        p = np.mod(np.arctan2(y, x) + np.pi, 2*np.pi) - np.pi
    elif on == 'eccentricity': p = np.sqrt(x**2 + y**2)
    elif on == 'rho':          p = np.sqrt(x**2 + y**2) * np.pi/180
    elif on == 'x':            p = x
    elif on == 'y':            p = y
    else: raise ValueError('unrecognized plot parameter: %s' % (on,))
    # okay, transform p:
    p = np.asarray(p if transform is None else transform(p))
    # if there are nans, let's get rid of them
    p[np.isnan(p)] = -np.inf
    # and run it through the cmap...
    clrs = cmap(p)
    # and set the appropriate pixels...
    im[ii] = clrs
    # check for undersize requests
    if boundary_pixels < 0:
        bp = -boundary_pixels
        whole_im = whole_im[bp:-bp, bp:-bp, :]
    # and return the whole thing!
    return whole_im

def to_rgba(val):
    '''
    to_rgba(val) is identical to matplotlib.colors.to_rgba(val) except that it operates over lists
      as well as individual elements to yield matrices of rgba values. In addition, it always yields
      numpy vectors or matrices.
    '''
    if pimms.is_npmatrix(val) and val.shape[1] == 4: return val
    try: return np.asarray(matplotlib.colors.to_rgba(val))
    except Exception: return np.asarray([matplotlib.colors.to_rgba(u) for u in val])
def color_overlap(color1, *args):
    '''
    color_overlap(color1, color2...) yields the rgba value associated with overlaying color2 on top
      of color1 followed by any additional colors (overlaid left to right). This respects alpha
      values when calculating the results.
    Note that colors may be lists of colors, in which case a matrix of RGBA values is yielded.
    '''
    args = list(args)
    args.insert(0, color1)
    rgba = np.asarray([0.5,0.5,0.5,0])
    for c in args:
        c = to_rgba(c)
        a = c[...,3]
        a0 = rgba[...,3]
        if   np.isclose(a0, 0).all(): rgba = np.ones(rgba.shape) * c
        elif np.isclose(a,  0).all(): continue
        else:                         rgba = times(a, c) + times(1-a, rgba)
    return rgba

_vertex_angle_empirical_prefixes = ['prf_', 'measured_', 'empiirical_']
_vertex_angle_model_prefixes = ['model_', 'predicted_', 'inferred_', 'template_', 'atlas_',
                                'benson14_', 'benson17_']
_vertex_angle_prefixes = ([''] + _vertex_angle_model_prefixes + _vertex_angle_model_prefixes)

def vertex_curvature_color(m):
    return [0.2,0.2,0.2,1.0] if m['curvature'] > -0.025 else [0.7,0.7,0.7,1.0]
def vertex_prop(m, property=None):
    if property is None or property is Ellipsis: return None
    elif pimms.is_vector(property): return property
    elif pimms.is_number(property): return property
    elif not pimms.is_str(property): raise ValueError('argument is not property-like: %s'%property)
    elif property in m: return m[property]
    else: return None
def vertex_weight(m, property=None):
    p = vertex_prop(m, property)
    if p is None:
        return next((m[k]
                     for name in ['weight', 'variance_explained']
                     for pref in ([''] + _vertex_angle_empirical_prefixes)
                     for k in [pref + name]
                     if k in m),
                    1.0)
    return p
def vertex_angle(m, property=None):
    p = vertex_prop(m, property)
    if p is None:
        ang0 = next((m[kk] for k in _vertex_angle_prefixes for kk in [k+'polar_angle'] if kk in m),
                    None)
        if ang0 is not None: return ang0
        ang0 = next((m[kk]
                     for name in ['angle', 'theta']
                     for k in _vertex_angle_prefixes for kk in [k + name]
                     if kk in m),
                    None)
        if ang0 is not None:
            return np.mod(90.0 - 180.0/np.pi*ang0 + 180, 360) - 180
        return None
    return p
def vertex_eccen(m, property=None):
    p = vertex_prop(m, property)
    if p is None:
        ecc0 = next((m[k]
                     for kk in _vertex_angle_prefixes
                     for k in [kk + 'eccentricity']
                     if k in m),
                    None)
        if ecc0 is not None: return ecc0
        ecc0 = next((m[k]
                     for kk in _vertex_angle_prefixes
                     for k in [kk + 'rho']
                     if k in m),
                    None)
        if ecc0 is not None:
            return 180.0/np.pi*ecc0
        return None
    return p
def vertex_sigma(m, property=None):
    p = vertex_prop(m, property)
    if p is not None: return p
    return next((m[k]
                 for kk in _vertex_angle_prefixes
                 for nm in ['sigma', 'radius', 'size']
                 for k in [kk + nm]
                 if k in m),
                None)
def vertex_varea(m, property=None):
    p = vertex_prop(m, property)
    if p is not None: return p
    return next((m[k]
                 for kk in _vertex_angle_prefixes
                 for nm in ['visual_area', 'varea', 'label']
                 for k in [kk + nm]
                 if k in m),
                None)
def vertex_angle_color(m, weight_min=0.2, weighted=True, hemi=None, property=Ellipsis,
                       weight=Ellipsis, null_color='curvature'):
    if m is Ellipsis:
        kw0 = {'weight_min':weight_min, 'weighted':weighted, 'hemi':hemi,
               'property':property, 'weight':weight, 'null_color':null_color}
        def _pass_fn(x, **kwargs):
            return vertex_angle_color(x, **{k:(kwargs[k] if k in kwargs else kw0[k])
                                            for k in list(set(kwargs.keys() + kw0.keys()))})
        return _pass_fn
    if isinstance(null_color, basestring):
        null_color = null_color.lower()
        if null_color == 'curvature' or null_color == 'curv':
            nullColor = np.asarray(vertex_curvature_color(m))
        else:
            raise ValueError('bad null color: %s' % null_color)
    else:
        nullColor = np.asarray(null_color)
    ang = vertex_angle(m, property)
    if ang is None: return nullColor
    if isinstance(hemi, basestring):
        hemi = hemi.lower()
        if hemi == 'lh' or hemi == 'left':
            ang = ang
        elif hemi == 'rh' or hemi == 'right':
            ang = -ang
        elif hemi == 'abs':
            ang = np.abs(ang)
        else: raise ValueError('bad hemi argument: %s' % hemi)
    w = vertex_weight(m, property=weight) if weighted else None
    if weighted and (not pimms.is_number(w) or w < weight_min):
        return nullColor
    angColor = np.asarray(cmap_polar_angle_sym((ang + 180.0) / 360.0))
    if weighted:
        return angColor*w + nullColor*(1-w)
    else:
        return angColor
def vertex_eccen_color(m, weight_min=0.1, weighted=True, hemi=None,
                       property=Ellipsis, null_color='curvature', weight=Ellipsis):
    if m is Ellipsis:
        kw0 = {'weight_min':weight_min, 'weighted':weighted, 'hemi':hemi,
               'property':property, 'weight':weight, 'null_color':null_color}
        def _pass_fn(x, **kwargs):
            return vertex_eccen_color(x, **{k:(kwargs[k] if k in kwargs else kw0[k])
                                            for k in list(set(kwargs.keys() + kw0.keys()))})
        return _pass_fn
    if isinstance(null_color, basestring):
        null_color = null_color.lower()
        if null_color == 'curvature' or null_color == 'curv':
            nullColor = np.asarray(vertex_curvature_color(m))
        else:
            raise ValueError('bad null color: %s' % null_color)
    else:
        nullColor = np.asarray(null_color)
    ecc = vertex_eccen(m, property)
    if ecc is None: return nullColor
    w = vertex_weight(m, property=weight) if weighted else None
    if weighted and (not pimms.is_number(w) or w < weight_min):
        return nullColor
    eccColor = np.asarray(cmap_eccentricity((ecc if 0 < ecc < 90 else 0 if ecc < 0 else 90)/90.0))
    if weighted:
        return eccColor*w + nullColor*(1-w)
    else:
        return eccColor
def vertex_sigma_color(m, weight_min=0.1, weighted=True, hemi=None,
                       property=Ellipsis, null_color='curvature', weight=Ellipsis):
    if m is Ellipsis:
        kw0 = {'weight_min':weight_min, 'weighted':weighted, 'hemi':hemi,
               'property':property, 'weight':weight, 'null_color':null_color}
        def _pass_fn(x, **kwargs):
            return vertex_sigma_color(x, **{k:(kwargs[k] if k in kwargs else kw0[k])
                                            for k in list(set(kwargs.keys() + kw0.keys()))})
        return _pass_fn
    if isinstance(null_color, basestring):
        null_color = null_color.lower()
        if null_color == 'curvature' or null_color == 'curv':
            nullColor = np.asarray(vertex_curvature_color(m))
        else:
            raise ValueError('bad null color: %s' % null_color)
    else:
        nullColor = np.asarray(null_color)
    sig = vertex_sigma(m, property)
    if sig is None: return nullColor
    w = vertex_weight(m, property=weight) if weighted else None
    if weighted and (not pimms.is_number(w) or w < weight_min):
        return nullColor
    sigColor = np.asarray(cmap_radius(sig/30.0))
    if weighted:
        return sigColor*w + nullColor*(1-w)
    else:
        return sigColor
_varea_colors = {'V1': (1,0,0), 'V2': (0,1,0),    'V3': (0,0,1),
                 'hV4':(0,1,1), 'VO1':(0,0.5,1), 'VO2':(0,1,0.5),
                 'LO1':(1,0,1), 'LO2':(0.5,0,1), 'TO1':(1,0,0.5), 'TO2':(0.5,0,0.5),
                 'V3a':(1,1,0), 'V3a':(0.5,1,0)}
def vertex_varea_color(m, property=Ellipsis, null_color='curvature', hemi=None,
                       weight=Ellipsis, weight_min=0.1, weighted=False):
    if m is Ellipsis:
        kw0 = {'weight_min':weight_min, 'weighted':weighted, 'hemi':hemi,
               'property':property, 'weight':weight, 'null_color':null_color}
        def _pass_fn(x, **kwargs):
            return vertex_varea_color(x, **{k:(kwargs[k] if k in kwargs else kw0[k])
                                            for k in list(set(kwargs.keys() + kw0.keys()))})
        return _pass_fn
    if isinstance(null_color, basestring):
        null_color = null_color.lower()
        if null_color == 'curvature' or null_color == 'curv':
            nullColor = np.asarray(vertex_curvature_color(m))
        else:
            raise ValueError('bad null color: %s' % null_color)
    else:
        nullColor = np.asarray(null_color)
    lbl = vertex_varea(m, property)
    if lbl is None or lbl == 0: return nullColor
    w = vertex_weight(m, property=weight) if weighted else None
    if weighted and (not pimms.is_number(w) or w < weight_min):
        return nullColor
    if not pimms.is_str(lbl):
        lbl = None if lbl is None or lbl < 0 or lbl > len(visual_area_names) else \
              visual_area_names[lbl]
    lblColor = np.asarray(_varea_colors.get(lbl, None))
    if lblColor is None: return nullColor
    if weighted: return lblColor*w + nullColor*(1-w)
    else:        return lblColor
def curvature_colors(m):
    '''
    curvature_colors(m) yields an array of curvature colors for the vertices of the given
      property-bearing object m.
    '''
    return np.asarray(m.map(vertex_curvature_color))
def retino_colors(vcolorfn, *args, **kwargs):
    'See eccen_colors, angle_colors, sigma_colors, and varea_colors.'
    if len(args) == 0:
        def _retino_color_pass(*args, **new_kwargs):
            return retino_colors(vcolorfn, *args,
                                 **{k:(new_kwargs[k] if k in new_kwargs else kwargs[k])
                                    for k in set(kwargs.keys() + new_kwargs.keys())})
        return _retino_color_pass
    elif len(args) > 1:
        raise ValueError('retinotopy color functions accepts at most one argument')
    m = args[0]
    # we need to handle the arguments
    if isinstance(m, (geo.VertexSet, pimms.ITable)):
        tbl = m.properties if isinstance(m, geo.VertexSet) else m
        n = tbl.row_count
        # if the weight or property arguments are lists, we need to thread these along
        if 'property' in kwargs:
            props = kwargs['property']
            del kwargs['property']
            if not (pimms.is_vector(props) or pimms.is_matrix(props)):
                props = [props for _ in range(n)]
        else: props = None
        if 'weight' in kwargs:
            ws = kwargs['weight']
            del kwargs['weight']
            if not pimms.is_vector(ws) and not pimms.is_matrix(ws): ws = [ws for _ in range(n)]
        else: ws = None
        vcolorfn0 = vcolorfn(Ellipsis, **kwargs) if len(kwargs) > 0 else vcolorfn
        if props is None and ws is None: vcfn = lambda m,k:vcolorfn0(m)
        elif props is None:              vcfn = lambda m,k:vcolorfn0(m, weight=ws[k])
        elif ws is None:                 vcfn = lambda m,k:vcolorfn0(m, property=props[k])
        else: vcfn = lambda m,k:vcolorfn0(m, property=props[k], weight=ws[k])
        return np.asarray([vcfn(r,kk) for (kk,r) in enumerate(tbl.rows)])
    else:
        return vcolorfn(m, **kwargs)
def angle_colors(*args, **kwargs):
    '''
    angle_colors(obj) yields an array of colors for the polar angle map of the given
      property-bearing object (cortex, tesselation, mesh).
    angle_colors(dict) yields an array of the color for the particular vertex property
      mapping that is given as dict.
    angle_colors() yields a functor version of angle_colors that can be called with one of the
      above arguments; note that this is useful precisely because the returned function
      preserves the arguments passed; e.g. angle_colors(weighted=False)(mesh) is equivalent to
      angle_colors(mesh, weighted=False).

    The following options are accepted:
      * weighted (True) specifies whether to use weight as opacity.
      * weight_min (0.2) specifies that below this weight value, the curvature (or null color)
        should be plotted.
      * property (Ellipsis) specifies the specific property that should be used as the
        polar angle value; if Ellipsis, will attempt to auto-detect this value.
      * weight (Ellipsis) specifies  the specific property that should be used as the weight value.
      * null_color ('curvature') specifies a color that should be used as the background.
    '''
    return retino_colors(vertex_angle_color, *args, **kwargs)
def eccen_colors(*args, **kwargs):
    '''
    eccen_colors(obj) yields an array of colors for the eccentricity map of the given
      property-bearing object (cortex, tesselation, mesh).
    eccen_colors(dict) yields an array of the color for the particular vertex property mapping
      that is given as dict.
    eccen_colors() yields a functor version of eccen_colors that can be called with one of the
      above arguments; note that this is useful precisely because the returned function
      preserves the arguments passed; e.g. eccen_colors(weighted=False)(mesh) is equivalent to
      eccen_colors(mesh, weighted=False).

    The following options are accepted:
      * weighted (True) specifies whether to use weight as opacity.
      * weight_min (0.2) specifies that below this weight value, the curvature (or null color)
        should be plotted.
      * property (Ellipsis) specifies the specific property that should be used as the
        eccentricity value; if Ellipsis, will attempt to auto-detect this value.
      * weight (Ellipsis) specifies  the specific property that should be used as the weight value.
      * null_color ('curvature') specifies a color that should be used as the background.
    '''
    return retino_colors(vertex_eccen_color, *args, **kwargs)
def sigma_colors(*args, **kwargs):
    '''
    sigma_colors(obj) yields an array of colors for the pRF-radius map of the given
      property-bearing object (cortex, tesselation, mesh).
    sigma_colors(dict) yields an array of the color for the particular vertex property mapping
      that is given as dict.
    sigma_colors() yields a functor version of sigma_colors that can be called with one of the
      above arguments; note that this is useful precisely because the returned function
      preserves the arguments passed; e.g. sigma_colors(weighted=False)(mesh) is equivalent to
      sigma_colors(mesh, weighted=False).

    Note: radius_colors() is an alias for sigma_colors().

    The following options are accepted:
      * weighted (True) specifies whether to use weight as opacity.
      * weight_min (0.2) specifies that below this weight value, the curvature (or null color)
        should be plotted.
      * property (Ellipsis) specifies the specific property that should be used as the
        eccentricity value; if Ellipsis, will attempt to auto-detect this value.
      * weight (Ellipsis) specifies  the specific property that should be used as the weight value.
      * null_color ('curvature') specifies a color that should be used as the background.
    '''
    return retino_colors(vertex_sigma_color, *args, **kwargs)
radius_colors = sigma_colors
def varea_colors(*args, **kwargs):
    '''
    varea_colors(obj) yields an array of colors for the visual area map of the given
      property-bearing object (cortex, tesselation, mesh).
    varea_colors(dict) yields an array of the color for the particular vertex property mapping
      that is given as dict.
    varea_colors() yields a functor version of varea_colors that can be called with one of the
      above arguments; note that this is useful precisely because the returned function
      preserves the arguments passed; e.g. varea_colors(weighted=False)(mesh) is equivalent to
      varea_colors(mesh, weighted=False).

    The following options are accepted:
      * weighted (True) specifies whether to use weight as opacity.
      * weight_min (0.2) specifies that below this weight value, the curvature (or null color)
        should be plotted.
      * property (Ellipsis) specifies the specific property that should be used as the
        eccentricity value; if Ellipsis, will attempt to auto-detect this value.
      * weight (Ellipsis) specifies  the specific property that should be used as the weight value.
      * null_color ('curvature') specifies a color that should be used as the background.
    '''
    return retino_colors(vertex_varea_color, *args, **kwargs)
def colors_to_cmap(colors):
    colors = np.asarray(colors)
    if len(colors.shape) == 1: return colors_to_cmap([colors])[0]
    if colors.shape[1] == 3:
        colors = np.hstack((colors, np.ones((len(colors),1))))
    steps = (0.5 + np.asarray(range(len(colors)-1), dtype=np.float))/(len(colors) - 1)
    return matplotlib.colors.LinearSegmentedColormap(
        'auto_cmap',
        {clrname: ([(0, col[0], col[0])] +
                   [(step, c0, c1) for (step,c0,c1) in zip(steps, col[:-1], col[1:])] +
                   [(1, col[-1], col[-1])])
         for (clridx,clrname) in enumerate(['red', 'green', 'blue', 'alpha'])
         for col in [colors[:,clridx]]},
        N=(len(colors)))
def guess_cortex_cmap(pname):
    '''
    guess_cortex_cmap(proptery_name) yields a tuple (cmap, (vmin, vmax)) of a cortical color map
      appropriate to the given property name and the suggested value scaling for the cmap. If the
      given property is not a string or is not recognized then the log_eccentricity axis is used
      and the suggested vmin and vmax are None.
    '''
    import matplotlib as mpl
    if isinstance(pname, mpl.colors.Colormap): pname = pname.name
    if not pimms.is_str(pname): return ('eccenflat', cmap_eccenflat, (None, None), None)
    if pname in colormaps: (cm,cmname) = (colormaps[pname],pname)
    else:
        # check each manually
        cm = None
        for (k,v) in six.iteritems(colormaps):
            if pname.endswith(k):
                (cmname,cm) = (k,v)
                break
        if cm is None:
            for (k,v) in six.iteritems(colormaps):
                if pname.startswith(k):
                    (cmname,cm) = (k,v)
                    break
    # we prefer log-eccentricity when possible
    if cm is None: return ('eccenflat', cmap_eccenflat, (None, None), None)
    if ('log_'+cmname) in colormaps:
        cmname = 'log_'+cmname
        cm = colormaps[cmname]
    return (cmname,) + (cm if len(cm) == 3 else cm + (None,))
def apply_cmap(zs, cmap, vmin=None, vmax=None, unit=None, logrescale=False):
    '''
    apply_cmap(z, cmap) applies the given cmap to the values in z; if vmin and/or vmax are passed,
      they are used to scale z.

    Note that this function can automatically rescale data into log-space if the colormap is a
    neuropythy log-space colormap such as log_eccentricity. To enable this behaviour use the
    optional argument logrescale=True.
    '''
    zs = pimms.mag(zs) if unit is None else pimms.mag(zs, unit)
    zs = np.asarray(zs, dtype='float')
    if pimms.is_str(cmap): cmap = matplotlib.cm.get_cmap(cmap)
    if logrescale:
        if vmin is None: vmin = np.log(np.nanmin(zs))
        if vmax is None: vmax = np.log(np.nanmax(zs))
        mn = np.exp(vmin)
        u = zdivide(nanlog(zs + mn) - vmin, vmax - vmin, null=np.nan)
    else:        
        if vmin is None: vmin = np.nanmin(zs)
        if vmax is None: vmax = np.nanmax(zs)
        u = zdivide(zs - vmin, vmax - vmin, null=np.nan)
    u[np.isnan(u)] = -np.inf
    return cmap(u)

def cortex_cmap_plot_2D(the_map, zs, cmap, vmin=None, vmax=None, axes=None, triangulation=None):
    '''
    cortex_cmap_plot_2D(map, zs, cmap, axes) plots the given cortical map values zs on the given
      axes using the given given color map and yields the resulting polygon collection object.
    cortex_cmap_plot_2D(map, zs, cmap) uses matplotlib.pyplot.gca() for the axes.

    The following options may be passed:
      * triangulation (None) may specify the triangularion object for the mesh if it has already
        been created; otherwise it is generated fresh.
      * axes (None) specify the axes on which to plot; if None, then matplotlib.pyplot.gca() is
        used. If Ellipsis, then a tuple (triangulation, z, cmap) is returned; to recreate the plot,
        one would call:
          axes.tripcolor(triangulation, z, cmap, shading='gouraud', vmin=vmin, vmax=vmax)
      * vmin (default: None) specifies the minimum value for scaling the property when one is passed
        as the color option. None means to use the min value of the property.
      * vmax (default: None) specifies the maximum value for scaling the property when one is passed
        as the color option. None means to use the max value of the property.
    '''
    if triangulation is None:
        triangulation = matplotlib.tri.Triangulation(the_map.coordinates[0], the_map.coordinates[1],
                                                     triangles=the_map.tess.indexed_faces.T)
    if axes is Ellipsis: return (triangulation, zs, cmap)
    return axes.tripcolor(triangulation, zs, cmap=cmap, shading='gouraud', vmin=vmin, vmax=vmax)
def cortex_rgba_plot_2D(the_map, rgba, axes=None, triangulation=None):
    '''
    cortex_rgba_plot_2D(map, rgba, axes) plots the given cortical map on the given axes using the
      given (n x 4) matrix of vertex colors and yields the resulting polygon collection object.
    cortex_rgba_plot_2D(map, rgba) uses matplotlib.pyplot.gca() for the axes.

    The option triangulation may also be passed if the triangularion object has already been
    created; otherwise it is generated fresh.
    '''
    cmap = colors_to_cmap(rgba)
    zs = np.linspace(0.0, 1.0, the_map.vertex_count)
    return cortex_cmap_plot_2D(the_map, zs, cmap, axes=axes, triangulation=triangulation)
def cortex_plot_colors(the_map,
                       color=None, cmap=None, vmin=None, vmax=None, alpha=None,
                       underlay='curvature', mask=None):
    '''
    cortex_plot_colors(mesh, opts...) yields the cortex colors as a matrix of RGBA rows for the
      given mesh and options. 

    The following options are accepted:
      * color (default: None) specifies the color to plot for each vertex; this argument may take a
        number of forms:
          * None, do not plot a color over the underlay (the default)
          * a matrix of RGB or RGBA values, one per vertex
          * a property vector or a string naming a property, in which case the cmap, vmin, and vmax
            arguments are used to generate colors
          * a function that, when passed a single argument, a dict of the properties of a single
            vertex, yields an RGB or RGBA list for that vertex.
      * cmap (default: 'eccenflat') specifies the colormap to use in plotting if the color
        argument provided is a property.
      * vmin (default: None) specifies the minimum value for scaling the property when one is passed
        as the color option. None means to use the min value of the property.
      * vmax (default: None) specifies the maximum value for scaling the property when one is passed
        as the color option. None means to use the max value of the property.
      * underlay (default: 'curvature') specifies the default underlay color to plot for the
        cortical surface; it may be None, 'curvature', or a color.
      * alpha (default None) specifies the alpha values to use for the color plot. If None, then
        leaves the alpha values from color unchanged. If a single number, then all alpha values in
        color are multiplied by that value. If a list of values, one per vertex, then this vector
        is multiplied by the alpha values. Finally, any negative value is set instead of multiplied.
        So, for example, if there were 3 vertices with:
          * color = ((0,0,0,1), (0,0,1,0.5), (0,0,0.75,0,8))
          * alpha = (-0.5, 1, 0.5)
        then the resulting colors plotted will be ((0,0,0,0.5), (0,0,1,0.5), (0,0,0.75,0,4)).
      * mask (default: None) specifies a mask to use for the mesh; this is passed through to_mask()
        to figure out the masking. Those vertices not in the mask are not plotted (but they will be
        plotted in the underlay if it is not None).
    '''
    # okay, let's interpret the color
    if color is None:
        color = np.full((the_map.vertex_count, 4), 0.5)
        color[:,3] = 0
    elif pimms.is_map(color) and len(color) == 1:
        (k,v) = next(six.iteritems(color))
        try: ktag = np.random.randint(sys.maxsize)
        except Exception: ktag = np.random.randint(sys.maxint)
        ktag = '%016x_%s' % (ktag, k)
        the_map = the_map.with_prop({ktag:v})
        return cortex_plot_colors(the_map, color=ktag,
                                  cmap=cmap, vmin=vmin, vmax=vmax, alpha=alpha,
                                  underlay=underlay, mask=mask)
    try:
        clr = matplotlib.colors.to_rgba(color)
        # This is an rgb color to plot...
        color = np.ones((the_map.vertex_count,4)) * matplotlib.colors.to_rgba(clr)
    except Exception: pass
    if pimms.is_vector(color) or pimms.is_str(color):
        # it's a property that gets interpreted via the colormap
        p = the_map.property(color)
        # if the colormap is none, we can try to guess it
        logtr = False
        if cmap is None:
            (cmapname,cmap,(vmn,vmx),unit) = guess_cortex_cmap(color)
            logtr = cmapname.startswith('log_')
        else:
            cmapname = cmap if pimms.is_str(cmap) else cmap.name
            if cmapname in colormaps:
                q = colormaps[cmapname]
                (cmap,(vmn,vmx),unit) = q if len(q) == 3 else (q + (None,))
            else:
                (vmn,vmx,unit) = (np.nanmin(p), np.nanmax(p), None)
        p = pimms.mag(p) if unit is None else pimms.mag(p, unit)
        if vmin is None: vmin = vmn
        if vmax is None: vmax = vmx
        # we use logrescale here because we assume that if color was 'eccentricity' or similar,
        # we want to rescale according to that colormap:
        color = apply_cmap(p, cmap, vmin=vmin, vmax=vmax, unit=unit, logrescale=logtr)
    if not pimms.is_matrix(color):
        # must be a function; let's try it...
        color = to_rgba(the_map.map(color))
    color = np.array(color)
    if color.shape[1] != 4: color = np.hstack([color, np.ones([color.shape[0], 1])])
    # okay, and the underlay...
    if underlay is not None:
        if pimms.is_str(underlay) and underlay.lower() in ['curvature', 'curv']:
            underlay = apply_cmap(the_map.prop('curvature'), cmap_curvature, vmin=-1, vmax=1)
        else:
            try: underlay = np.ones((the_map.vertex_count, 4)) * to_rgba(underlay)
            except Exception: raise ValueError('plot underlay failed: must be a color or curvature')
    # okay, let's check on alpha...
    if alpha is not None:
        if pimms.is_number(alpha): alpha = np.full(color.shape[0], alpha)
        else: alpha = the_map.property(alpha)
        color[:,3] *= alpha
        neg = (alpha < 0)
        color[neg,3] = -alpha[neg]
    alpha = color[:,3]
    # and the mask...
    if mask is not None:
        ii = the_map.mask(mask, indices=True)
        tmp = np.zeros(len(color))
        tmp[ii] = color[ii,3]
        color[:,3] = tmp
    # then, blend with the underlay if need be
    if underlay is not None:
        color = color_overlap(underlay, color)
    return color

def cortex_plot_2D(the_map,
                   color=None, cmap=None, vmin=None, vmax=None, alpha=None,
                   underlay='curvature', mask=None, axes=None, triangulation=None):
    '''
    cortex_plot_2D(map) yields a plot of the given 2D cortical mesh, map.

    The following options are accepted:
      * color (default: None) specifies the color to plot for each vertex; this argument may take a
        number of forms:
          * None, do not plot a color over the underlay (the default)
          * a matrix of RGB or RGBA values, one per vertex
          * a property vector or a string naming a property, in which case the cmap, vmin, and vmax
            arguments are used to generate colors
          * a function that, when passed a single argument, a dict of the properties of a single
            vertex, yields an RGB or RGBA list for that vertex.
      * cmap (default: 'eccenflat') specifies the colormap to use in plotting if the color
        argument provided is a property.
      * vmin (default: None) specifies the minimum value for scaling the property when one is passed
        as the color option. None means to use the min value of the property.
      * vmax (default: None) specifies the maximum value for scaling the property when one is passed
        as the color option. None means to use the max value of the property.
      * underlay (default: 'curvature') specifies the default underlay color to plot for the
        cortical surface; it may be None, 'curvature', or a color.
      * alpha (default None) specifies the alpha values to use for the color plot. If None, then
        leaves the alpha values from color unchanged. If a single number, then all alpha values in
        color are multiplied by that value. If a list of values, one per vertex, then this vector
        is multiplied by the alpha values. Finally, any negative value is set instead of multiplied.
        So, for example, if there were 3 vertices with:
          * color = ((0,0,0,1), (0,0,1,0.5), (0,0,0.75,0,8))
          * alpha = (-0.5, 1, 0.5)
        then the resulting colors plotted will be ((0,0,0,0.5), (0,0,1,0.5), (0,0,0.75,0,4)).
      * mask (default: None) specifies a mask to use for the mesh; thi sis passed through map.mask()
        to figure out the masking. Those vertices not in the mask are not plotted (but they will be
        plotted in the underlay if it is not None).
      * axes (default: None) specifies a particular set of matplotlib pyplot axes that should be
        used. If axes is Ellipsis, then instead of attempting to render the plot, a tuple of
        (tri, zs, cmap) is returned; in this case, tri is a matplotlib.tri.Triangulation
        object for the given map and zs and cmap are an array and colormap (respectively) that
        will produce the correct colors. Without axes equal to Ellipsis, these would instead
        be rendered as axes.tripcolor(tri, zs, cmap, shading='gouraud'). If axes is None, then
        uses the current axes.
      * triangulation (default: None) specifies the matplotlib triangulation object to use, if one
        already exists; otherwise a new one is made.
    '''
    # parse the axes
    if axes is None: axes = matplotlib.pyplot.gca()
    # process the colors
    color = cortex_plot_colors(the_map, color=color, cmap=cmap, vmin=vmin, vmax=vmax, alpha=alpha,
                               underlay=underlay, mask=mask)
    # finally, we can make the plot!
    return cortex_rgba_plot_2D(the_map, color, axes=axes, triangulation=triangulation)


# 3D Graphics ######################################################################################

# If we're using Python 2, we're compatible with pysurfer:
def _ipyvolume_load_error(*args, **kwargs):
    raise RuntimeError('load failure: the requested object could not be loaded, probably ' +
                       'because you do not have ipyvolume installed correctly')
cortex_plot_3D = _ipyvolume_load_error
try:
    import ipyvolume as ipv

    def cortex_plot_3D(obj,
                       color=None, cmap=None, vmin=None, vmax=None, alpha=None,
                       underlay='curvature', mask=None, hemi=None, surface='inflated',
                       figure=Ellipsis, width=600, height=600, mesh_alpha=None,
                       view=None, camera_distance=100, camera_fov=None, camera_up=None):
        '''
    cortex_plot_3D(hemi) plots the inflated surface of the given cortex object hemi and returns the
      ipyvolume figure object.
    cortex_plot_3D(mesh) plots the given mesh.
    cortex_plot_3D((hemi1, hemi2...)) plots all the given hemispheres or meshes.
    cortex_plot_3D(mesh) yields a PySurfer Brain object for the given 3D cortical mesh.
    cortex_plot_3D(subject) is equivalent to cortex_plot_3D((subject.lh, subject.rh)).

    The following options are accepted:
      * color (default: None) specifies the color to plot for each vertex; this argument may take a
        number of forms:
          * None, do not plot a color over the underlay (the default)
          * a matrix of RGB or RGBA values, one per vertex
          * a property vector or a string naming a property, in which case the cmap, vmin, and vmax
            arguments are used to generate colors
          * a function that, when passed a single argument, a dict of the properties of a single
            vertex, yields an RGB or RGBA list for that vertex.
      * cmap (default: 'eccenflat') specifies the colormap to use in plotting if the color
        argument provided is a property.
      * vmin (default: None) specifies the minimum value for scaling the property when one is passed
        as the color option. None means to use the min value of the property.
      * vmax (default: None) specifies the maximum value for scaling the property when one is passed
        as the color option. None means to use the max value of the property.
      * underlay (default: 'curvature') specifies the default underlay color to plot for the
        cortical surface; it may be None, 'curvature', or a color.
      * alpha (default None) specifies the alpha values to use for the color plot. If None, then
        leaves the alpha values from color unchanged. If a single number, then all alpha values in
        color are multiplied by that value. If a list of values, one per vertex, then this vector
        is multiplied by the alpha values. Finally, any negative value is set instead of multiplied.
        So, for example, if there were 3 vertices with:
          * color = ((0,0,0,1), (0,0,1,0.5), (0,0,0.75,0,8))
          * alpha = (-0.5, 1, 0.5)
        then the resulting colors plotted will be ((0,0,0,0.5), (0,0,1,0.5), (0,0,0.75,0,4)).
      * mesh_alpha (default: None) specifies the alpha to use for the mesh object itself. This may
        be single value (0 for transparent, 1 for opaque) or a vector of values, one per vertex, or
        a property name.
      * mask (default: None) specifies a mask to use for the mesh; this is passed through to_mask()
        to figure out the masking. Those vertices not in the mask are not plotted (but they will be
        plotted in the underlay if it is not None).
      * hemi (defaut: None) specifies the hemisphere to use. If the passed mesh object is actually a
        subject or mesh pair then this specifies which hemisphere to use. If the passed object is a
        mesh, then this overrides its chirality, if specified in meta_data. If two hemispheres are
        given, then this may be 'both' or 'split' in accordinace with PySurfer's Brain() class.
      * surface (default: 'white') specifies the surface to use if the mesh object passed is in fact
        either a cortex or subject object.
      * axes (default: None) specifies the ipyvolume axes that should be used; if None, then the
        current axes are used.
      * stylize (default: True) specifies whether or not neuropythy should apply additional standard
        stylings to the ipyvolume figure; these include making sure that the plot range is
        appropriate, that the axes and box are not plotted, and that the camera is in an
        orthographic position looking at the mesh.
    '''
        # First, see if we've been passed a cortex or subject object...
        mesh = []
        for arg in (obj if pimms.is_vector(obj) else [obj]):
            if   geo.is_mesh(arg): mesh.append(arg)
            elif geo.is_topo(arg): mesh.append(geo.to_mesh((arg, surface)))
            elif mri.is_subject(arg):
                for h in (hemi if pimms.is_vector(hemi) else [hemi]):
                    if geo.is_topo(h) or geo.is_mesh(h): hh = [h]
                    elif h in arg.hemis: hh = [arg.hemis[h]]
                    else:
                        h = to_hemi_str(h)
                        hh = [arg.lh, arg.rh] if h == 'lr' else [arg.hemis[h]]
                    for hhh in hh: mesh.append(geo.to_mesh((hhh, surface)))
        # process the colors
        rgba = np.concatenate(
            [cortex_plot_colors(m, color=color, cmap=cmap, vmin=vmin, vmax=vmax,
                                alpha=alpha, underlay=underlay, mask=mask)
             for m in mesh])
        n = len(rgba)
        # process the mesh_alpha parameter
        if mesh_alpha is None:
            mesh_alpha = [None for m in mesh]
        elif pimms.is_scalar(mesh_alpha):
            if mesh_alpha == 1: mesh_alpha = None
            elif pimms.is_str(mesh_alpha): mesh_alpha = [m.prop(mesh_alpha) for m in mesh]
            else: mesh_alpha = [mesh_alpha for m in mesh]
        elif len(mesh_alpha) != len(mesh):
            mesh_alpha = [mesh_alpha for m in mesh]
        # Okay, setup the ipyv figure...
        mns = np.full(3, np.inf)
        mxs = np.full(3, -np.inf)
        ms = ()
        if figure is None: f = ipv.gcf()
        elif figure is Ellipsis: f = ipv.figure(width=width, height=height)
        else: f = figure
        i0 = 0
        for (m,ma) in zip(mesh, mesh_alpha):
            (x,y,z) = m.coordinates
            ii = slice(i0, i0 + m.vertex_count)
            i0 += m.vertex_count
            ipvm = ipv.plot_trisurf(x,y,z, m.tess.indexed_faces.T, color=rgba[ii,:3])
            mns = np.nanmin([mns, np.nanmin([x,y,z], axis=1)], axis=0)
            mxs = np.nanmax([mxs, np.nanmax([x,y,z], axis=1)], axis=0)
            ms = ms + (ipvm,)
            # handle mesh alpha, if given
            if ma is not None:
                if pimms.is_scalar(ma): ma = np.full([m.vertex_count, 1], ma)
                else: ma = np.reshape(ma, [m.vertex_count, 1])
                ipvm.color = np.hstack([ipvm.color, ma])
                ipvm.material.transparent = True
        # Figure out the bounding box...
        szs = mxs - mns
        mid = 0.5*(mxs + mns)
        bsz = np.nanmax(szs)
        # okay, set the plot limits
        mxs = mid + 0.5*bsz
        mns = mid - 0.5*bsz
        ipv.pylab.xlim(mns[0],mxs[0])
        ipv.pylab.ylim(mns[1],mxs[1])
        ipv.pylab.zlim(mns[2],mxs[2])
        # few other styling things
        ipv.pylab.xlabel('')
        ipv.pylab.ylabel('')
        ipv.pylab.zlabel('')
        ipv.style.box_off()
        ipv.style.axes_off()
        # figure out the view
        d = camera_distance
        up = None
        if view is None: view = 'back'
        if pimms.is_str(view):
            if camera_distance is None: d = 10
            view = view.lower()
            (view, up) = (((0,-d,0), (0,0, 1)) if view in ['back','rear','posterior','p','-y'] else
                          ((0, d,0), (0,0, 1)) if view in ['front','anterior','a','+y','y']    else
                          ((0,0, d), (0, 1,0)) if view in ['top','superior','s','+z','z']      else
                          ((0,0,-d), (0,-1,0)) if view in ['bottom','inferior','i','-z']       else
                          (( d,0,0), (0,0, 1)) if view in ['right','r','x','+x']               else
                          ((-d,0,0), (0,0, 1)) if view in ['left','l','-x']                    else
                          (view, None))
            if pimms.is_str(view): raise ValueError('Unknown view: %s' % view)
        if camera_up is not None: up = camera_up
        if d is None: d = np.sqrt(np.sum(np.asarray(d)**2))
        fov = camera_fov
        if fov is None: fov = 180.0/np.pi * 2 * np.arctan(1.125/(2*d))
        f.camera.position = tuple(view)
        f.camera.up = tuple(up)
        f.camera.fov = fov
        f.camera.lookAt(tuple(mid))
        warnings.warn('neuropythy: NOTE: due to a bug in ipyvolume, camera views cannot currently' +
                      ' be set by neuropythy; however, if you click the reset (home) button in' +
                      ' the upper-left corner of the figure, the requested view will be fixed.')
        return f
except Exception: pass

def cortex_plot(mesh, *args, **opts):
    '''
    cortex_plot(mesh) calls either cortex_plot_2D or cortex_plot_3D depending on the dimensionality
      of the given mesh, and yields the resulting graphics object. All optional arguments supported
      by each is supported by cortex plot.

    The following options are accepted:
      * color (default: None) specifies the color to plot for each vertex; this argument may take a
        number of forms:
          * None, do not plot a color over the underlay (the default)
          * a matrix of RGB or RGBA values, one per vertex
          * a property vector or a string naming a property, in which case the cmap, vmin, and vmax
            arguments are used to generate colors
          * a function that, when passed a single argument, a dict of the properties of a single
            vertex, yields an RGB or RGBA list for that vertex.
      * cmap (default: 'eccenflat') specifies the colormap to use in plotting if the color
        argument provided is a property.
      * vmin (default: None) specifies the minimum value for scaling the property when one is passed
        as the color option. None means to use the min value of the property.
      * vmax (default: None) specifies the maximum value for scaling the property when one is passed
        as the color option. None means to use the max value of the property.
      * underlay (default: 'curvature') specifies the default underlay color to plot for the
        cortical surface; it may be None, 'curvature', or a color.
      * alpha (default None) specifies the alpha values to use for the color plot. If None, then
        leaves the alpha values from color unchanged. If a single number, then all alpha values in
        color are multiplied by that value. If a list of values, one per vertex, then this vector
        is multiplied by the alpha values. Finally, any negative value is set instead of multiplied.
        So, for example, if there were 3 vertices with:
          * color = ((0,0,0,1), (0,0,1,0.5), (0,0,0.75,0,8))
          * alpha = (-0.5, 1, 0.5)
        then the resulting colors plotted will be ((0,0,0,0.5), (0,0,1,0.5), (0,0,0.75,0,4)).
      * mask (default: None) specifies a mask to use for the mesh; thi sis passed through map.mask()
        to figure out the masking. Those vertices not in the mask are not plotted (but they will be
        plotted in the underlay if it is not None).
      * hemi (defaut: None) specifies the hemisphere to use. If the passed mesh object is actually a
        subject or mesh pair then this specifies which hemisphere to use. If the passed object is a
        mesh, then this overrides its chirality, if specified in meta_data. If two hemispheres are
        given, then this may be 'both' or 'split' in accordinace with PySurfer's Brain() class.
      * surface (default: 'white') specifies the surface to use if the mesh object passed is in fact
        either a cortex or subject object.
      * axes (default: None) specifies a particular set of matplotlib pyplot axes that should be
        used. If axes is Ellipsis, then instead of attempting to render the plot, a tuple of
        (tri, zs, cmap) is returned; in this case, tri is a matplotlib.tri.Triangulation
        object for the given map and zs and cmap are an array and colormap (respectively) that
        will produce the correct colors. Without axes equal to Ellipsis, these would instead
        be rendered as axes.tripcolor(tri, zs, cmap, shading='gouraud'). If axes is None, then
        uses the current axes.
      * triangulation (default: None) specifies the matplotlib triangulation object to use, if one
        already exists; otherwise a new one is made.
    '''
    if not isinstance(mesh, geo.Mesh) or mesh.coordinates.shape[0] > 2:
        # must be a 3D call
        return cortex_plot_3D(mesh, *args, **opts)
    else:
        return cortex_plot_2D(mesh, *args, **opts)

class ROIDrawer:
    '''
    ROIDrawer(axes, mproj) creates a new ROIDrawer class that interacts with the plot on
      the given axes to save the lines drawn on the plot as a neuropythy path trace, which
      is available from the roiDrawer object as roiDrawer.trace once the user is finished.
    ROIDrawer(axes, fmap) extracts the map projection from the meta-data in the given
      flatmap; note that if the projection is not encoded in the flatmap's meta-data, an
      error will be raised.
    
    All arguments and keyword arguments following the first two are passed verbatim to the
    axes.plot() function, which is used for plotting the drawn lines.
    '''
    def __init__(self, axes, mp, closed=True, event_handlers=None, plot_list=None, *args, **kw):
        from neuropythy import (to_map_projection, is_map_projection, is_vset)
        # Assumption: the axes have already been plotted, so there's no need to do any
        # plotting here, aside from the lines we're about to draw
        self.axes = axes
        # get the map projection
        if is_vset(mp) and 'projection' in mp.meta_data: mp = mp.meta_data['projection']
        elif not is_map_projection(mp): mp = to_map_projection(mp)
        self.map_projection = mp
        # draw the initial lines
        if len(args) == 0 and len(kw) == 0: args = ['k.-']
        self.line = axes.plot([], [], *args, **kw)[0]
        self.xs = list(self.line.get_xdata())
        self.ys = list(self.line.get_ydata())
        self.connections = [
            self.line.figure.canvas.mpl_connect('button_press_event', self.on_button),
            self.line.figure.canvas.mpl_connect('key_press_event', self.on_key),
            self.line.figure.canvas.mpl_connect('close_event', self.on_close)]
        # get rid of the mesh for path traces (because we don't want to export meshes on accident)
        if mp.mesh is not None: mp = mp.copy(mesh=None)
        self.trace = geo.PathTrace(mp, np.zeros((2,0)), closed=closed,
                                   meta_data={'roi_drawer':self})
        self.closed = closed
        self.event_handlers = event_handlers
        self.plot_list = plot_list
        if plot_list is not None and len(plot_list) > 0:
            for p in plot_list[1:]: p.set_visible(False)
            plot_list[0].set_visible(True)
        self.current_plot = 0
    def end(self, success=True):
        from neuropythy import path_trace
        # we've finished; clean up and make the line
        if success:
            if len(self.xs) < 1: raise ValueError('Drawn line has no points')
            pts = np.transpose([self.xs, self.ys])
            # remove us from the trace meta-data if we're in it
            rd = self.trace.meta_data.get('roi_drawer')
            if rd is self: self.trace.meta_data = self.trace.meta_data.discard('roi_drawer')
            self.trace.persist()
        else: self.trace = None
        if self.line:
            for conn in self.connections:
                self.line.figure.canvas.mpl_disconnect(conn)
        # redraw the final version:
        if self.closed:
            self.xs.append(self.xs[0])
            self.ys.append(self.ys[0])
            self.line.set_data(self.xs, self.ys)
            self.line.figure.canvas.draw()
        matplotlib.pyplot.close(self.line.figure)
        # clear everything
        self.connection = None
        self.line = None
        self.xs = None
        self.ys = None
    def on_close(self, event):
        self.end(success=False)
    def on_key(self, event):
        import matplotlib.collections
        if event.inaxes != self.line.axes: return
        if event.key == 'tab':
            # we go to the next plot...
            if not self.plot_list: return
            self.plot_list[self.current_plot].set_visible(False)
            tmp = self.current_plot
            self.current_plot = (self.current_plot + 1) % len(self.plot_list)
            a = self.plot_list[self.current_plot]
            if isinstance(a, matplotlib.collections.TriMesh): a.set_visible(True)
            else: self.plot_list[self.current_plot] = a(self.axes)
            a.figure.canvas.draw()
    def on_button(self, event):
        endkeys = ['shift+control', 'shift+ctrl', 'control+shift', 'ctrl+shift']
        if self.line is None: return
        if event.inaxes != self.line.axes: return
        # if shift is down, we delete the last point
        if event.key == 'shift':
            if len(self.xs) == 0: return
            self.xs = self.xs[:-1]
            self.ys = self.ys[:-1]
        elif event.key in endkeys:
            # we abort!
            self.end(success=False)
            return
        elif (event.key is not None and self.event_handlers is not None and
              event.key in self.event_handlers):
            self.event_handlers[event.key](self, event)
            return
        else: # add the points
            self.xs.append(event.xdata)
            self.ys.append(event.ydata)
        # redraw the line regardless
        self.line.set_data(self.xs, self.ys)
        self.line.figure.canvas.draw()
        # and update the trace
        self.trace.points = np.asarray([self.xs, self.ys])
        if event.dblclick or event.key in ['control', 'ctrl']:
            self.end(success=True)

def trace_roi(hemi, map_proj, axes, closed=True, event_handlers=None, plot_list=None, **kw):
    '''
    trace_roi(hemi, map_proj, axes) creates an ROIDrawer object that controls the tracing of lines
      around an ROI in a 2D matplotlib plot and returns a not-yet-persistent immutable PathTrace
      object with the ROIDrawer in its meta_data. The path trace is persisted as soon as the user
      finished drawing their line; if the line is canceled, then the trace is never persisted.

    ROI tracing is very simple: any point in the plot is appended to the path as it is clicked; in
    order to eliminate the previous point, hold shift while clicking. To end the path, hold control
    while clicking. To abort the path, hold both shift and control while clicking. (Double-clicking
    should be equivalent to control-clicking, but this does not work in all setups.) In order to use
    the ROI tracing, `%matplotlib notebook` is recommended.

    The trace_roi() function accepts all options that can be passed to cortex_plot() as well as the
    following options:
      * closed (default: True) specifies whether the path-trace that is constructed should be closed
        (True) or open (False).
      * event_handlers (default: None) specifies additional event handlers (named by key) for the
        ROIDrawer().
      * plot_list (default: None) specifies a list of alternate TriMesh objects that can be plotted
        cyclically when the user presses tab. TriMesh objects can be created by pyplot.triplot and
        pyplot.tripcolor, which are used by the neuropythy cortex_plot function as well. If the
        plot_list is not empty, then the first item of the list is immediately plotted on the axes.
        Unlike in the ROIDrawer function itself, the plot_list may contain maps whose keys are
        the various arguments (aside from the initial mesh argument) to cortex_plot.
    '''
    # okay, first off, if the plot_list has maps in it, we convert them using cortex_plot:
    if plot_list is not None:
        if geo.is_flatmap(hemi):                  fmap = hemi
        elif geo.is_flatmap(map_proj):            fmap = map_proj
        elif not geo.is_map_projection(map_proj): fmap = geo.to_map_projection(map_proj)(hemi)
        else:                                     fmap = map_proj(hemi)
        plot_list = [cortex_plot(fmap, axes=axes, **p) if pimms.is_map(p) else p
                     for p in plot_list]
    # next, make the roi drawer
    rd = ROIDrawer(axes, map_proj, closed=closed,
                   event_handlers=event_handlers, plot_list=plot_list)
    return rd.trace

        



####################################################################################################
# commands/register_retinotopy.py
# The code for the function that handles the registration of retinotopy
# By Noah C. Benson

from __future__ import print_function

import numpy                        as     np
import scipy                        as     sp
import nibabel                      as     nib
import nibabel.freesurfer.io        as     fsio
import nibabel.freesurfer.mghformat as     fsmgh
import pyrsistent                   as     pyr
import os, sys, six, pimms

from ..vision                       import (register_retinotopy, retinotopy_model, clean_retinotopy,
                                            empirical_retinotopy_data)
from ..util                         import (nanlt, nangt)

info = '''
The register_retinotopy command can be used to register a subject's
hemisphere(s) to a model of retinotopy in the early visual cortex. At least one 
subject id (either a freesurfer subject name, if SUBJECTS_DIR is set
appropriately in the environment, or a path to a subject directory) must be
given. Registration to a retinotopic model is performed for both hemispheres of
all of these subjects.

This process of registration is fundamentally Bayesian in that it finds a
compromise between prior belief about retinotopy (the model of retinotopy) and
the observed retinotopic measurements of the subject.

In each subject's freesurfer directory, a variety of output data is deposited:
 * surf/lh.retinotopy.sphere.reg
   surf/rh.retinotopy.sphere.reg
   These files contain the registrations of the left and right hemispheres to
   the retinotopy model. They are in the same format as Freesurfer's other 
   surface-data files such as surf/lh.white.
 * surf/lh.inferred_angle   surf/rh.inferred_angle
   surf/lh.inferred_eccen   surf/rh.inferred_eccen
   surf/lh.inferred_varea   surf/rh.inferred_varea
   surf/lh.inferred_sigma   surf/rh.inferred_sigma
   These files contain predictions of polar angle, eccentricity, visual-area
   label, and pRF radius (sigma) for each hemisphere. The files are in
   freesurfer's 'curv' or 'morphology' format (same as lh.curv); though they
   may be exported in other formats.
 * mri/inferred_angle.mgz
   mri/inferred_eccen.mgz
   mri/inferred_varea.mgz
   mri/inferred_sigma.mgz
   These contain the data from the above surface data projected into the
   subject's 3D volume. Note that the volumes are oriented like Freesurfer's
   mri/brain.mgz file; if you want to convert this to the orientation of your
   original anatomical scan, use mri_convert:
    > mri_convert -rl mri/rawavg.mgz mri/angle_predict.mgz \\
                  mri/scanner.angle_predict.mgz

The following options are accepted:
 * --help|-h
   Prints this message.
 * --verbose|-v
   Specifies that progress messages should be printed during the registration.
 * --lh-eccen=|-e<file>
   --lh-angle=|-a<file>
   --lh-weight=|-w<file>
   --lh-radius=|-q<file>
   --lh-theta=|-t<file>
   --lh-rho=|-r<file>
   --rh-eccen=|-e<file>
   --rh-angle=|-a<file>
   --rh-weight=|-w<file>
   --rh-radius=|-q<file>
   --rh-theta=|-t<file>
   --rh-rho=|-r<file>
   Each of these arguments specifies the name of a data file to load in as a
   representation of the subject's eccentricity (eccen), polar angle (angle), 
   pRF radius (radius), or weight; these should be the names of either an
    mgh/mgz files whose size is 1 x 1 x n, where n is the number of vertices in
   the hemisphere for the subject, or a FreeSurfer curv-style filename with n
   vertices. By default, files in the subject's surf directory that match a
   name-template are automatically loaded and used. This template is name
   <hemi>.<tag><name>, optionally ending with .mgz, where tag is one of (and in
   order of preference) 'prf_', 'empirical_', 'measured_', 'training_', or '',
   and name is one of 'eccentricity'/'eccen', 'polar_angle'/'angle',
   'radius'/'prfsz'/'sigma', or 'weight'/'variance_explained'/'vexpl'. Note that
   the options ?h-theta and ?h-rho specify that the polar angle (theta) or
   eccentricity (rho) are in radians.
 * --no-volume-export|-x
   --no-surface-export|-z
   --no-registration-export|-X
   Specifies that either volumes, surfaces, or the registration (.sphere.reg)
   files should not be exported.
 * --no-overwrite|-n
   Specifies that files should not be overwritten if they exist already. 
 * --no-lh|-k
   --no-rh|-K
   Specifies that the given hemisphere should not be registered.
 * --clean|-c
   Instructs the algorithm to clean the retinotopy with a basic smoothing
   routine before performing registration (EXPERIMENTAL).
 * --model-sym|-S
   Specifies that the model used should be a version of the Schira2010 model
   as used in Benson et al. (2014) PLOS Comput Biol; this model is on the
   fsaverage_benson17 pseudo-hemisphere.
 * --no-resample|-b
   Indicates that the cortical map should not be resampled to a uniform
   triangle grid immediately prior to registration. This is generally only a
   good idea when used when the subject being registered is the fsaverage or
   fsaverage_sym subject.
 * --no-invert-rh-angle|-V
   If the RH hemisphere polar angle values do not need to be negated in order to
   match the model, this option should be given. Note that both the standard and
   the fsaverage_sym model (see --model-sym option) encode the polar angle
   values as ranging from 0 degrees (upper vertical meridian) to 180 degrees
   (lower vertical meridian). Therefore, this option should generally not be
   used unless your RH data has already been inverted.
 * --weight-min=|-m<value>
   The cutoff value to use for the weight; 0.1 by default. Weights less than
   this will be truncated to 0.
 * --scale=|-s<weight>
   Specifies the strength of the functional forces relative to anatomical forces
   during the registration; higher values will generally result in more warping
   while lower values will result in less warping. The default value is 20.
 * --field-sign-weight=|-g<weight>
   --radius-weight=|-G<weight>
   These specify the strength of the field-sign-based or radius-based matching
   techniques. Both the field-sign and the pRF radius are used to modify the
   weights on the individual vertices prior to the registration. Values of 0
   indicate that this part of the weight should be ignored while values of 1
   indicate that the weight should be relatively strong (default is 1 for both).
 * --max-steps=|-i<steps>
   This option specifies the maximum number of steps to run the registration; by
   default this is 2000.
 * --max-step-size=|-D<value>
   This specifies the max step-size for any single vertex; by default this is
   0.05.
 * --prior=|-p<name>
   This specifies the name of the prior registration to use in the fsaverage or
   fsaverage_sym subject; by default this is 'benson17'.
 * --eccen-tag=|-y<tag>
   --angle-tag=|-t<tag>
   --label-tag=|-l<tag>
   --radius-tag=|-j<tag>
   These options specify the output tag to use for the predicted measurement
   that results from the registration. By default, these are
   'inferred_eccen', 'inferred_angle', 'inferred_varea', and 'inferred_sigma'.
   The output files have the name <hemi>.<tag>
 * --registration-name=|-u<string>
   This parameter indicates that the registration file, by default named 
   lh.retinotopy.sphere.reg, should instead be named lh.<string>.sphere.reg.
 * --max-output-eccen=|-M<val>
   This specifies the maximum eccentricity to include in the output; there is no
   particular need to limit one's output, but it can be done with this argument.
   By default this is 90.
 * --max-input-eccen=|-I<val>
   --min-input-eccen=|-J<val>
   This specifies the minimum/maximum eccentricity to include in the
   registration. By default these are 90 and 0 degrees, but it is generally a
   good idea to exclude eccentricity values above those used in the stimulus
   presentation.
 * --no-volume-export|-x
   --no-surface-export|-z
   --no-registration-export|-X
   These flags indicate that the various data produced and written to the
   filesystem under normal execution should be suppressed. The volume export
   refers to the predicted volume files exported to the subject's mri directory;
   the registration export refers to the <hemi>.retinotopy_sym.sphere.reg file,
   written to the subject's surf directory, that contains the registered
   coordinates for the subject; and the surface export refers to the
   <hemi>.eccen_predict.mgz and similar files that are written to the
   subject's surf directory.
 * --surf-format=|-f<fmt>
   Specifies that the surface files should be exported in the given format. By
   default this is curv (FreeSurfer curvature files, such as lh.curv), but
   mgh, mgz, or nifti are all valid options.
 * --vol-format=|-F<fmt>
   Specifies that the volume files should be exported in the given format. By
   default this is mgz, but mgh or nifti are also valid options.
 * --surf-outdir=|-o<dir>
   --vol-outdir=|-O<dir>
   Specifies the output directories for the surface and volume files; by default
   this uses the subject's FreeSurfer directory + /surf or /mri.
 * --subjects-dir=|-d
   Specifies additional subject directory search locations (in addition to the
   SUBJECTS_DIR environment variable and the FREESURFER_HOME/subjects
   directories, which are given here in descending search priority) when looking
   for subjects by name. This option cannot be specified multiple times, but it
   may contain : characters to separate directories, as in PATH.
 * --
   This token, by itself, indicates that the arguments that remain should not be
   processed as flags or options, even if they begin with a -.
'''
_retinotopy_parser_instructions = [
    # Flags
    ('h', 'help',                   'help',              False),
    ('v', 'verbose',                'verbose',           False),
    ('x', 'no-volume-export',       'no_vol_export',     False),
    ('z', 'no-surface-export',      'no_surf_export',    False),
    ('X', 'no-registration-export', 'no_reg_export',     False),
    ('n', 'no-overwrite',           'no_overwrite',      False),
    ('k', 'no-lh',                  'run_lh',            True),
    ('K', 'no-rh',                  'run_rh',            True),
    ('c', 'clean',                  'clean',             False),
    ('b', 'no-resample',            'resample',          True),
    ('N', 'partial-correction',     'part_vol_correct',  False),
    ('S', 'model-sym',              'model_sym',         False),
    ('V', 'no-invert-rh-angle',     'invert_rh_angle',   True),
    # Options
    ['a', 'lh-angle',               'angle_lh_file',     None],
    ['t', 'lh-theta',               'theta_lh_file',     None],
    ['e', 'lh-eccen',               'eccen_lh_file',     None],
    ['r', 'lh-rho',                 'rho_lh_file',       None],
    ['w', 'lh-weight',              'weight_lh_file',    None],
    ['q', 'lh-radius',              'radius_lh_file',    None],
    ['A', 'rh-angle',               'angle_rh_file',     None],
    ['T', 'rh-theta',               'theta_rh_file',     None],
    ['E', 'rh-eccen',               'eccen_rh_file',     None],
    ['R', 'rh-rho',                 'rho_rh_file',       None],
    ['W', 'rh-weight',              'weight_rh_file',    None],
    ['Q', 'rh-radius',              'radius_rh_file',    None],

    ['m', 'weight-min',             'weight_min',        '0.1'],
    ['s', 'scale',                  'scale',             '20.0'],
    ['g', 'field-sign-weight',      'field_sign_weight', '1.0'],
    ['G', 'radius-weight',          'radius_weight',     '1.0'],
    ['i', 'max-steps',              'max_steps',         '2500'],
    ['D', 'max-step-size',          'max_step_size',     '0.02'],
    ['p', 'prior',                  'prior',             'benson17'],

    ['f', 'surf-format',            'surface_format',    'curv'],
    ['F', 'vol-format',             'volume_format',     'mgz'],
    ['o', 'surf-outdir',            'surface_path',      None],
    ['O', 'vol-outdir',             'volume_path',       None],
    ['y', 'eccen-tag',              'eccen_tag',         'inferred_eccen'],
    ['w', 'angle-tag',              'angle_tag',         'inferred_angle'],
    ['l', 'label-tag',              'label_tag',         'inferred_varea'],
    ['j', 'radius-tag',             'radius_tag',        'inferred_sigma'],
    ['u', 'registration-name',      'registration_name', 'retinotopy'],
    ['M', 'max-output-eccen',       'max_out_eccen',     '90'],
    ['I', 'max-input-eccen',        'max_in_eccen',      '90'],
    ['J', 'min-input-eccen',        'min_in_eccen',      '0'],
    ['d', 'subjects-dir',           'subjects_dir',      None]]
_retinotopy_parser = pimms.argv_parser(_retinotopy_parser_instructions)

def _guess_surf_file(fl):
    # MGH/MGZ files
    try: return np.asarray(fsmgh.load(fl).dataobj).flatten()
    except Exception: pass
    # FreeSurfer Curv files
    try: return fsio.read_morph_data(fl)
    except Exception: pass
    # Nifti files
    try: return np.squeeze(nib.load(fl).dataobj)
    except Exception: pass
    raise ValueError('Could not determine filetype for: %s' % fl)
def _guess_vol_file(fl):
    # MGH/MGZ files
    try: return fsmgh.load(fl)
    except Exception: pass
    # Nifti Files
    try: return nib.load(fl)
    except Exception: pass
    raise ValueError('Could not determine filetype for: %s' % fl)

@pimms.calc('subject', 'model', 'options', 'note', 'error',
            'no_vol_export',     
            'no_surf_export',    
            'no_reg_export',     
            'no_overwrite',
            'model_sym',
            'invert_rh_angle',
            'part_vol_correct',  
            'angle_lh_file',     
            'theta_lh_file',     
            'eccen_lh_file',     
            'rho_lh_file',
            'run_lh',
            'weight_lh_file',
            'radius_lh_file',
            'angle_rh_file',     
            'theta_rh_file',     
            'eccen_rh_file',     
            'rho_rh_file',
            'run_rh',
            'clean',
            'weight_rh_file',
            'radius_rh_file',
            'weight_min',        
            'scale',             
            'max_steps',         
            'max_step_size',     
            'prior',             
            'surface_format',    
            'volume_format',     
            'surface_path',      
            'volume_path',       
            'eccen_tag',         
            'angle_tag',         
            'label_tag',
            'radius_tag',
            'registration_name', 
            'max_out_eccen',
            'max_in_eccen',
            'min_in_eccen',
            'resample',
            'field_sign_weight',
            'radius_weight')
def calc_arguments(args):
    '''
    calc_arguments is a calculator that parses the command-line arguments for the registration
    command and produces the subject, the model, the log function, and the additional options.
    '''
    (args, opts) = _retinotopy_parser(args)
    # We do some of the options right here...
    if opts['help']:
        print(info, file=sys.stdout)
        sys.exit(1)
    # and if we are verbose, lets setup a note function
    verbose = opts['verbose']
    def note(s):
        if verbose:
            print(s, file=sys.stdout)
            sys.stdout.flush()
        return verbose
    def error(s):
        print(s, file=sys.stderr)
        sys.stderr.flush()
        sys.exit(1)
    if len(args) < 1: error('subject argument is required')
    try: # we try FreeSurfer first:
        import neuropythy.freesurfer as fs
        # Add the subjects directory, if there is one
        if 'subjects_dir' in opts and opts['subjects_dir'] is not None:
            fs.add_subject_path(opts['subjects_dir'])
        # Get the subject now
        sub = fs.subject(args[0])
    except Exception: sub = None
    if sub is None:
        try: # As an alternative, try HCP
            import neuropythy.hcp as hcp
            # Add the subjects directory, if there is one
            if 'subjects_dir' in opts and opts['subjects_dir'] is not None:
                hcp.add_subject_path(opts['subjects_dir'])
            sub = hcp.subject(args[0])
        except Exception: sub = None
    if sub is None: error('Failed to load subject %s' % args[0])
    # and the model
    if len(args) > 1:       mdl_name = args[1]
    elif opts['model_sym']: mdl_name = 'schira'
    else:                   mdl_name = 'benson17'
    try:
        if opts['model_sym']:
            model = {h:retinotopy_model(mdl_name).persist() for h in ['lh', 'rh']}
        else:
            model = {h:retinotopy_model(mdl_name, hemi=h).persist() for h in ['lh', 'rh']}
    except Exception: error('Could not load retinotopy model %s' % mdl_name)

    # Now, we want to run a few filters on the options
    # Parse the simple numbers
    for o in ['weight_min', 'scale', 'max_step_size', 'max_out_eccen',
              'max_in_eccen', 'min_in_eccen', 'field_sign_weight', 'radius_weight']:
        opts[o] = float(opts[o])
    opts['max_steps'] = int(opts['max_steps'])
    # Make a note:
    note('Processing subject: %s' % sub.name)
    del opts['help']
    del opts['verbose']
    del opts['subjects_dir']
    # That's all we need!
    return pimms.merge(opts,
                       {'subject': sub.persist(),
                        'model':   pyr.pmap(model),
                        'options': pyr.pmap(opts),
                        'note':    note,
                        'error':   error})
@pimms.calc('cortices')
def calc_retinotopy(note, error, subject, clean, run_lh, run_rh,
                    invert_rh_angle, max_in_eccen, min_in_eccen,
                    angle_lh_file, theta_lh_file,
                    eccen_lh_file, rho_lh_file,
                    weight_lh_file, radius_lh_file,
                    angle_rh_file, theta_rh_file,
                    eccen_rh_file, rho_rh_file,
                    weight_rh_file, radius_rh_file):
    '''
    calc_retinotopy extracts the retinotopy options from the command line, loads the relevant files,
    and stores them as properties on the subject's lh and rh cortices.
    '''
    ctcs = {}
    for (h,ang,tht,ecc,rho,wgt,rad,run) in [
            ('lh', angle_lh_file,theta_lh_file, eccen_lh_file,rho_lh_file,
             weight_lh_file, radius_lh_file, run_lh),
            ('rh', angle_rh_file,theta_rh_file, eccen_rh_file,rho_rh_file,
             weight_rh_file, radius_rh_file, run_rh)]:
        if not run: continue
        hemi = getattr(subject, h)
        props = {}
        # load the properties or find them in the auto-properties
        if ang:
            try:
                props['polar_angle'] = _guess_surf_file(ang)
            except Exception: error('could not load surface file %s' % ang)
        elif tht:
            try:
                tmp = _guess_surf_file(tht)
                props['polar_angle'] = 90.0 - 180.0 / np.pi * tmp
            except Exception: error('could not load surface file %s' % tht)
        else:
            props['polar_angle'] = empirical_retinotopy_data(hemi, 'polar_angle')
        if ecc:
            try:
                props['eccentricity'] = _guess_surf_file(ecc)
            except Exception: error('could not load surface file %s' % ecc)
        elif rho:
            try:
                tmp = _guess_surf_file(rhp)
                props['eccentricity'] = 180.0 / np.pi * tmp
            except Exception: error('could not load surface file %s' % rho)
        else:
            props['eccentricity'] = empirical_retinotopy_data(hemi, 'eccentricity')
        if wgt:
            try: props['weight'] = _guess_surf_file(wgt)
            except Exception: error('could not load surface file %s' % wgt)
        else:
            props['weight'] = empirical_retinotopy_data(hemi, 'weight')
        if rad:
            try: props['radius'] = _guess_surf_file(rad)
            except Exception: error('could not load surface file %s' % rad)
        else:
            props['radius'] = empirical_retinotopy_data(hemi, 'radius')
        # Check for inverted rh
        if h == 'rh' and invert_rh_angle:
            props['polar_angle'] = -props['polar_angle']
        # and zero-out weights for high eccentricities
        props['weight'] = np.array(props['weight'])
        if max_in_eccen is not None:
            props['weight'][nangt(props['eccentricity'], max_in_eccen)] = 0
        if min_in_eccen is not None:
            props['weight'][nanlt(props['eccentricity'], min_in_eccen)] = 0
        # Do smoothing, if requested
        if clean:
            note('Cleaning %s retinotopy...' % h.upper())
            (ang,ecc) = clean_retinotopy(hemi, retinotopy=props, mask=None, weight='weight')
            props['polar_angle']  = ang
            props['eccentricity'] = ecc
        ctcs[h] = hemi.with_prop(props)
    return {'cortices': pyr.pmap(ctcs)}
@pimms.calc('registrations')
def calc_registrations(note, error, cortices, model, model_sym,
                       weight_min, scale, prior, max_out_eccen, max_steps, max_step_size,
                       radius_weight, field_sign_weight, resample, invert_rh_angle,
                       part_vol_correct):
    '''
    calc_registrations is the calculator that performs the registrations for the left and right
    hemisphere; these are returned as the immutable maps yielded from the register_retinotopy
    command.
    '''
    rsamp = ('fsaverage_sym' if model_sym else 'fsaverage') if resample else False
    # Do the registration
    res = {}
    for (h,ctx) in six.iteritems(cortices):
        note('Preparing %s Registration...' % h.upper())
        try:
            res[h] = register_retinotopy(ctx, model[h],
                                         model_hemi='sym' if model_sym else h,
                                         polar_angle='polar_angle',
                                         eccentricity='eccentricity',
                                         weight='weight',
                                         weight_min=weight_min,
                                         partial_voluming_correction=part_vol_correct,
                                         field_sign_weight=field_sign_weight,
                                         radius_weight=radius_weight,
                                         scale=scale,
                                         prior=prior,
                                         resample=rsamp,
                                         invert_rh_field_sign=invert_rh_angle,
                                         max_steps=max_steps,
                                         max_step_size=max_step_size,
                                         yield_imap=True)
        except Exception: #error('Exception caught while setting-up register_retinotopy (%s)' % h)
            raise
    return {'registrations': pyr.pmap(res)}
@pimms.calc('surface_files')
def save_surface_files(note, error, registrations, subject,
                       no_surf_export, no_reg_export, surface_format, surface_path,
                       angle_tag, eccen_tag, label_tag, radius_tag, registration_name):
    '''
    save_surface_files is the calculator that saves the registration data out as surface files,
    which are put back in the registration as the value 'surface_files'.
    '''
    if no_surf_export: return {'surface_files': ()}
    surface_format = surface_format.lower()
    # make an exporter for properties:
    if surface_format in ['curv', 'morph', 'auto', 'automatic']:
        def export(flnm, p):
            fsio.write_morph_data(flnm, p)
            return flnm
    elif surface_format in ['mgh', 'mgz']:
        def export(flnm, p):
            flnm = flnm + '.' + surface_format
            dt = np.int32 if np.issubdtype(p.dtype, np.dtype(int).type) else np.float32
            img = fsmgh.MGHImage(np.asarray([[p]], dtype=dt), np.eye(4))
            img.to_filename(flnm)
            return flnm
    elif surface_format in ['nifti', 'nii', 'niigz', 'nii.gz']:
        surface_format = 'nii' if surface_format == 'nii' else 'nii.gz'
        def export(flnm, p):
            flnm = flnm + '.' + surface_format
            dt = np.int32 if np.issubdtype(p.dtype, np.dtype(int).type) else np.float32
            img = nib.Nifti1Image(np.asarray([[p]], dtype=dt), np.eye(4))
            img.to_filename(flnm)
            return flnm
    else:
        error('Could not understand surface file-format %s' % surface_format)
    path = surface_path if surface_path else os.path.join(subject.path, 'surf')
    files = []
    note('Exporting files...')
    for h in six.iterkeys(registrations):
        hemi = subject.hemis[h]
        reg = registrations[h]
        note('Extracting %s predicted mesh...' % h.upper())
        pmesh = reg['predicted_mesh']
        for (pname,tag) in zip(['polar_angle', 'eccentricity', 'visual_area', 'radius'],
                               [angle_tag, eccen_tag, label_tag, radius_tag]):
            flnm = export(os.path.join(path, h + '.' + tag), pmesh.prop(pname))
            files.append(flnm)
        # last do the registration itself
        if registration_name and not no_reg_export:
            flnm = os.path.join(path, h + '.' + registration_name + '.sphere.reg')
            fsio.write_geometry(flnm, pmesh.coordinates.T, pmesh.tess.faces.T)
    return {'surface_files': tuple(files)}
def get_subject_image(subject):
    try: im = subject.images['brain']
    except Exception: im = None
    if im is None:
        try: im = subject.images['ribbon']
        except Exception: im = None
    if im is None:
        try: im = subject.images['original']
        except Exception: im = None
    if im is None:
        raise ValueError('Subject %s is missing files: no known images' % (subject.path,))
    return im
def get_subject_affine(subject):
    im = get_subject_image(subject)
    return im.affine
@pimms.calc('volume_files')
def save_volume_files(note, error, registrations, subject,
                      no_vol_export, volume_format, volume_path,
                      angle_tag, eccen_tag, label_tag, radius_tag):
    '''
    save_volume_files is the calculator that saves the registration data out as volume files,
    which are put back in the registration as the value 'volume_files'.
    '''
    from neuropythy.mri import image_clear
    if no_vol_export: return {'volume_files': ()}
    volume_format = volume_format.lower()
    # make an exporter for properties:
    affmatrix = get_subject_affine(subject)
    if volume_format in ['mgh', 'mgz', 'auto', 'automatic', 'default']:
        volume_format = 'mgh' if volume_format == 'mgh' else 'mgz'
        def export(flnm, d):
            if not pimms.is_nparray(d): d = np.asarray(d.dataobj)
            flnm = flnm + '.' + volume_format
            dt = np.int32 if np.issubdtype(d.dtype, np.dtype(int).type) else np.float32
            img = fsmgh.MGHImage(np.asarray(d, dtype=dt), affmatrix)
            img.to_filename(flnm)
            return flnm
    elif volume_format in ['nifti', 'nii', 'niigz', 'nii.gz']:
        volume_format = 'nii' if volume_format == 'nii' else 'nii.gz'
        def export(flnm, p):
            if not pimms.is_nparray(p): p = np.asarray(p.dataobj)
            flnm = flnm + '.' + volume_format
            dt = np.int32 if np.issubdtype(p.dtype, np.dtype(int).type) else np.float32
            img = nib.Nifti1Image(np.asarray(p, dtype=dt), affmatrix)
            img.to_filename(flnm)
            return flnm
    else:
        error('Could not understand volume file-format %s' % volume_format)
    path = volume_path if volume_path else os.path.join(subject.path, 'mri')
    files = []
    note('Extracting predicted meshes for volume export...')
    hemis = [registrations[h]['predicted_mesh'] if h in registrations else None
             for h in ['lh', 'rh']]
    note('Addressing image...')
    im = get_subject_image(subject)
    addr = (subject.lh.image_address(im), subject.rh.image_address(im))
    for (pname,tag) in zip(['polar_angle', 'eccentricity', 'visual_area', 'radius'],
                           [angle_tag, eccen_tag, label_tag, radius_tag]):
        # we have to make the volume first...
        dat = tuple([None if h is None else h.prop(pname) for h in hemis])
        (mtd,dt) = ('nearest',np.int32) if pname == 'visual_area' else ('linear',np.float32)
        note('Constructing %s image...' % pname)
        img = subject.cortex_to_image(dat, image_clear(im),
                                      method=mtd, dtype=dt, address=addr)
        flnm = export(os.path.join(path, tag), img)
        files.append(flnm)
    return {'volume_files': tuple(files)}
@pimms.calc('files')
def accumulate_files(surface_files, volume_files):
    '''
    accumulate_files is a calculator that just accumulates the exported files into a single tuple,
    files.
    '''
    return {'files': (tuple(surface_files) + tuple(volume_files))}

register_retinotopy_plan = pimms.plan(args=calc_arguments,
                                      retinotopy=calc_retinotopy,
                                      register=calc_registrations,
                                      surf_export=save_surface_files,
                                      vol_export=save_volume_files,
                                      accum_files=accumulate_files)

def main(args):
    '''
    register_retinotopy.main(args) can be given a list of arguments, such as sys.argv[1:]; these
    arguments may include any options and must include at least one subject id. All subjects whose
    ids are given are registered to a retinotopy model, and the resulting registration, as well as
    the predictions made by the model in the registration, are exported.
    '''
    m = register_retinotopy_plan(args=args)
    # force completion
    files = m['files']
    if len(files) > 0:
        return 0
    else:
        print('Error: No files exported.', file=sys.stderr)
        return 1




####################################################################################################
# neuropythy/commands/retinotopy.py
# The code for the function that handle the processing of retinotopy data from the command-line.
# By Noah C. Benson

from __future__ import print_function

import os, sys, six, re, pimms, textwrap, warnings
import numpy                        as     np
import scipy                        as     sp
import nibabel                      as     nib
import nibabel.freesurfer.io        as     fsio
import nibabel.freesurfer.mghformat as     fsmgh
import pyrsistent                   as     pyr

from   ..freesurfer                 import (subject as freesurfer_subject)
from   ..hcp                        import (subject as hcp_subject)
from   ..util                       import (curry, library_path, AutoDict, auto_dict)
from   ..util.conf                  import config
from   ..                           import io as nyio

# Note that much of this code has been stolen from atlas.py; at some point in the future, this
# should all be separated out into a set of classes/functions that handle the common operations.
@pimms.calc('worklog')
def calc_worklog(stdout=Ellipsis, stderr=Ellipsis, verbose=False):
    '''
    calc_worklog constructs the worklog from the stdout, stderr, stdin, and verbose arguments.
    '''
    try: cols = int(os.environ['COLUMNS'])
    except Exception: cols = 80
    return pimms.worklog(columns=cols, stdout=stdout, stderr=stderr, verbose=verbose)
@pimms.calc('subject')
def calc_subject(argv, worklog):
    '''
    calc_subject converts a subject_id into a subject object.

    Afferent parameters:
      @ argv
        The FreeSurfer subject name(s), HCP subject ID(s), or path(s) of the subject(s) to which the
        atlas should be applied.
    '''
    if len(argv) == 0: raise ValueError('No subject-id given')
    elif len(argv) > 1: worklog.warn('WARNING: Unused subject arguments: %s' % (argv[1:],))
    subject_id = argv[0]
    try:
        sub = freesurfer_subject(subject_id)
        if sub is not None:
            worklog('Using FreeSurfer subject: %s' % sub.path)
            return sub
    except Exception: pass
    try:
        sub = hcp_subject(subject_id)
        if sub is not None:
            worklog('Using HCP subject: %s' % sub.path)
            return sub
    except Exception: pass
    raise ValueError('Could not load subject %s' % subject_id)
@pimms.calc('hemisphere_data', 'hemisphere_tags')
def calc_hemispheres(subject, hemispheres='lh,rh'):
    '''
    calc_hemispheres extracts the relevant hemispheres from the subject object.

    Afferent parameters:
      @ hemispheres 
        The names of the hemispheres to be used in the retinotopy calculations. The default value
        is "lh,rh", but other hemisphere names, such as "lh_LR32k,rh_LR32k" (for an HCP-pipeline
        subject) can be given. All hemispheres should be separated by commas. If you want to give
        a hemisphere a particular tag, you can specify it as tag:hemi, e.g.:
        "lh:lh_LR32k,rh:rh_LR32k"; tags are used with other options for specifying data specific
        to a hemisphere.
    '''
    if not hemispheres: hemispheres = 'lh,rh'
    # first, separate by commas
    hs = hemispheres.split(',')
    hemis = {}
    hlist = []
    for h in hs:
        (tag,name) = h.split(':') if ':' in h else (h,h)
        h = subject.hemis.get(name, None)
        if h is None:
            raise ValueError('Give subject does not have a hemisphere named "%s"' % (name,))
        hemis[tag] = h
        hlist.append(tag)
    return {'hemisphere_data': pyr.pmap(hemis),
            'hemisphere_tags': hlist}
@pimms.calc('label_data')
def calc_labels(subject, hemisphere_tags, hemisphere_data, labels=None):
    '''
    calc_labels finds the available label data for the subject on which the retinotopy operations
    are being performed.

    Afferent parameters:
      @ labels 
        The filenames of the files containing label data for the subject's hemispheres. Label data
        can be provided in mgz, annot, or curv files containing visual area labels, one per vertex.
        The labels argument may be specified as a comma-separated list of filenames (in the same
        order as the hemispheres, which are lh then rh by default) or as a single template filename
        that may contain the character * as a stand-in for the hemisphere tag. For example,
        '/data/*.v123_labels.mgz' would look for the file /data/lh.v123_labels.mgz for the 'lh'
        hemisphere and for /data/rh_LR32k.v123_labels.mgz for the 'rh_LR32k' hemisphere.
        Note that labels are not required--without labels, no field-sign minimization is performed,
        so retinotopic cleaning may be less reliable. Note that additionally, labels may be
        preceded by the relevant tag; so instead of '/data/*.v123_labels.mgz' with, as in the 
        example, hemispheres 'lh,rh_LR32k', one could use the arguments
        'lh:/data/lh.v123_labels.mgz,rh:/data/rh_LR32k.v123_labels.mgz' (for labels) and
        'lh,rh:rh_LR32k' for hemispheres.
    '''
    lbls = {}
    # no argument this is fine--no labels are used
    if lbls is None: return {'label_data': pyr.m()}
    if not pimms.is_str(labels): raise ValueError('could not understand non-string labels')
    # first, it might just be a template pattern
    fls = {}
    if '*' in labels:
        sparts = labels.split('*')
        for h in hemisphere_tags:
            flnm = h.join(sparts)
            fls[h] = os.path.expanduser(os.path.expandvars(flnm))
    else:
        # okay, separate commas...
        lsplit = labels.split(',')
        for (k,l) in enumerate(lsplit):
            if ':' in l: (tag,name) = l.split(':')
            elif k < len(hemisphere_tags): (tag,name) = (hemisphere_taks[k],l)
            else: raise ValueError('could not match labels to hemispheres')
            if tag not in hemisphere_data:
                raise ValueError('Tag %s (in labels arg) does not exist' % (tag,))
            fls[tag] = os.path.expanduser(os.path.expandvars(name))
    for (tag,name) in six.iteritems(fls):
        if not os.path.isfile(name):
            raise ValueError('Labels filename %s not found' % (name,))
        hem = hemisphere_data[tag]
        tmp = nyio.load(name)
        if not pimms.is_vector(tmp) or len(tmp) != hem.vertex_count:
            raise ValueError('Labels file %s does not contain label data' % (name,))
        lbls[tag] = np.asarray(tmp)
    return {'label_data': pimms.persist(lbls)}
@pimms.calc('raw_retinotopy')
def calc_retinotopy(hemisphere_data, hemisphere_tags, label_data,
                    angles='*.prf_angle.mgz',
                    eccens='*.prf_eccen.mgz',
                    weights='*.prf_vexpl.mgz'):
    '''
    calc_retinotopy imports the raw retinotopy data for the given subject.

    Afferent parameters:
      @ angles 
        The filenames of the polar-angle files that are needed for each hemisphere. For more 
        information on how these files are specified, see the help text for the labels parameter.
        If angles is not supplied, then the default value is '*.prf_angle.mgz'. Polar angles
        MUST be encoded in clockwise degrees of rotation starting from the positive y-axis.
      @ eccens 
        The filenames of the eccentricity files that are needed for each hemisphere. For more 
        information on how these files are specified, see the help text for the labels parameter.
        If eccens is not supplied, then the default value is '*.prf_eccen.mgz'. Eccentricity
        MUST be encoded in degrees of visual angle.
      @ weights 
        The filenames of the weights (usually fraction of variance explained) files that are needed
        for each hemisphere. For more innformation on how these files are specified, see the help
        text for the labels parameters. If eccens is not supplied, then the default value is
        '*.prf_vexpl.mgz'. Variance explained should be encoded as a fraction with 1 indicating
        100% variance explained.
    '''
    retino = {}
    for (k,val) in zip(['angle', 'eccen', 'weight'], [angles, eccens, weights]):
        if not pimms.is_str(val): raise ValueError('could not understand non-string %ss' % k)
        # first, it might just be a template pattern
        fls = {}
        if '*' in val:
            sparts = val.split('*')
            for h in hemisphere_tags:
                flnm = h.join(sparts)
                fls[h] = os.path.expanduser(os.path.expandvars(flnm))
        else:
            # okay, separate commas...
            lsplit = val.split(',')
            for (kk,l) in enumerate(lsplit):
                if ':' in l: (tag,name) = l.split(':')
                elif kk < len(hemisphere_tags): (tag,name) = (hemisphere_tags[kk],l)
                else: raise ValueError('could not match %ss to hemispheres' % (k,))
                if tag not in hemisphere_data:
                    raise ValueError('Tag %s (in %ss arg) does not exist' % (tag,k))
                fls[tag] = os.path.expanduser(os.path.expandvars(name))
        retino[k] = fls
    # now go through and load them
    res = {}
    for (k,fls) in six.iteritems(retino):
        rr = {}
        for (tag,name) in six.iteritems(fls):
            if not os.path.isfile(name):
                raise ValueError('%ss filename %s not found' % (k,name,))
            hem = hemisphere_data[tag]
            tmp = nyio.load(name)
            if not pimms.is_vector(tmp) or len(tmp) != hem.vertex_count:
                raise ValueError('%ss file %s does not contain label data' % (k,name,))
            rr[tag] = np.asarray(tmp)
        res[k] = rr
    return {'raw_retinotopy': pimms.persist(res)}
@pimms.calc('clean_retinotopy')
def calc_clean_maps(raw_retinotopy, hemisphere_data, label_data, worklog,
                    no_clean=False):
    '''
    calc_clean_maps calculates cleaned retintopic maps.

    Afferent parameters:
      @ no_clean 
        May be set to True to indicate that cleaning of the retinotopic maps should be skipped;
        downstream calculations such as that of cortical magnification will instead be performed on
        the raw and not the clean retinotopy.
    '''
    from neuropythy.vision import (clean_retinotopy, as_retinotopy)
    if no_clean:
        worklog('Skipping retinotopic map cleaning...')
        cl = {h:{'polar_angle': raw_retinotopy['angle'][h],
                 'eccentricity': raw_retinotopy['eccen'][h],
                 'visual_area': label_data[h]}
              for h in six.iterkeys(hemisphere_data)}
        return {'clean_retinotopy': pimms.persist(cl)}
    worklog('Calculating cleaned retinotopic maps...')
    wl = worklog.indent()
    res = {}
    for (h,hem) in six.iteritems(hemisphere_data):
        wl('%s' % h)
        # collect the retinotopy and run the cleaning
        ret = {'polar_angle':        raw_retinotopy['angle'][h],
               'eccentricity':       raw_retinotopy['eccen'][h],
               'variance_explained': raw_retinotopy['weight'][h]}
        if label_data and h in label_data: ret['visual_area'] = label_data[h]
        cl = clean_retinotopy(hem, retinotopy=ret)
        res[h] = {'polar_angle': cl[0], 'eccentricity': cl[1]}
        if label_data and h in label_data: res[h]['visual_area'] = label_data[h]
    # That's all...
    return {'clean_retinotopy': pimms.persist(res)}
@pimms.calc('spotlight_cmag')
def calc_spotlight_cmag(hemisphere_data, clean_retinotopy, worklog,
                        no_spotlight=False, surface='midgray', nnearest=0.1):
    '''
    calc_spotlight_cmag calculates the spotlight-based cortical magnification data for the given
    retinotopic maps, after cleaning if not skipped.

    Afferent parameters:
      @ no_spotlight 
        May be used to indicate that no spotlight search should be performed.
      @ surface 
        Specifies the surface (white, pial, midgray) on which to calculate the cotical
        magnification; by default this is midgray.
      @ nnearest 
        Specifies the number of vertices to include in the nearest-neighbor calculation used for
        spotlight cortical magnification. If not specified, then 10% of vertices are used. This
        should be a number f where 0 <= f < 1 (a fraction of the total vertices) or a number of
        vertices; the former is encouaged.
    '''
    from neuropythy.vision import (areal_cmag, as_retinotopy)
    if no_spotlight:
        worklog('Skipping spotlight cortical magnification calculation.')
        return {'spotlight_cmag': None}
    worklog('Calculating spotlight cortical magnification...')
    wl = worklog.indent()
    res = {}
    for (h,cl) in six.iteritems(clean_retinotopy):
        hem = hemisphere_data[h]
        lbls = cl.get('visual_area', None)
        if lbls is None:
            wl('%s: skipping (no label data provided)' % (h,))
            continue
        else:
            wl('%s' % (h,))
        hh = np.zeros(hem.vertex_count)
        for l in np.unique(lbls):
            if l == 0: continue
            ii = np.where(lbls == l)[0]
            nn = nnearest if nnearest >= 1 else int(np.ceil(nnearest * len(ii)))
            cm = areal_cmag(hem, cl, surface_area=(surface+'_surface_area'), mask=ii, nnearest=nn)
            xy = np.transpose(as_retinotopy(cl, 'geographical'))
            hh[ii] = cm(xy[ii])
        res[h] = hh
    return {'spotlight_cmag': res}
@pimms.calc('radtan_cmag')
def calc_radtan_cmag(hemisphere_data, clean_retinotopy, worklog,
                     no_radtan=False, surface='midgray'):
    '''
    calc_radtan_cmag calculates the radial/tangential cortical magnification data for the given
    retinotopic maps, after cleaning if not skipped.

    Afferent parameters:
      @ no_radtan 
        May be used to indicate that no radial/tangential cortical magnification should be
        calculated.
      @ surface 
        Specifies the surface (white, pial, midgray) on which to calculate the cotical
        magnification; by default this is midgray.
    '''
    from neuropythy.vision import (disk_vmag, as_retinotopy)
    if no_radtan:
        worklog('Skipping radial/tangential cortical magnification calculation.')
        return {'radtan_cmag': None}
    worklog('Calculating radial and tangential cortical magnification...')
    wl = worklog.indent()
    res = {}
    for (h,hem) in six.iteritems(hemisphere_data):
        hem = hemisphere_data[h]
        cl = clean_retinotopy[h]
        lbls = cl.get('visual_area', None)
        if lbls is None:
            wl('%s: WARNING: Calculating without labels' % (h,))
            res[h] = disk_vmag(hem, cl)
        else:
            wl('%s' % (h,))
            hh = np.zeros((hem.vertex_count, 2))
            for l in np.unique(lbls):
                if l == 0: continue
                ii = np.where(lbls == l)[0]
                cm = disk_vmag(hem, cl, surface=surface, mask=ii)
                hh[ii,:] = cm[ii]
        res[h] = hh
    return {'radtan_cmag': res}
@pimms.calc('filemap', 'export_all_fn')
def calc_filemap(subject, worklog, radtan_cmag, spotlight_cmag, clean_retinotopy, no_clean=False,
                 output_path=None, overwrite=False, output_format='mgz', create_directory=False):
    '''
    calc_filemap is a calculator that prepares the calculated cortical magnification data and the
    cleaned retinotopic maps for exporting.

    Afferent parameters
      @ output_path 
        The directory into which the atlas files should be written. If not provided or None then
        uses the subject's surf directory. If this directory doesn't exist, then it uses the
        subject's directory itself.
      @ overwrite 
        Whether to overwrite existing atlas files. If True, then atlas files that already exist will
        be overwritten. If False, then no files are overwritten.
      @ create_directory 
        Whether to create the output path if it doesn't exist. This is False by default.
      @ output_format 
        The desired output format of the files to be written. May be one of the following: 'mgz',
        'mgh', or either 'curv' or 'morph'.

    Efferent values:
      @ filemap 
        A pimms lazy map whose keys are filenames and whose values are interpolated atlas
        properties.
      @ export_all_fn 
        A function of no arguments that, when called, exports all of the files in the filemap to the
        output_path.
    '''
    if output_path is None:
        output_path = os.path.join(subject.path, 'surf')
        if not os.path.isdir(output_path): output_path = subject.path
    output_format = 'mgz' if output_format is None else output_format.lower()
    if output_format.startswith('.'): output_format = output_format[1:]
    (fmt,ending) = (('mgh','.mgz') if output_format == 'mgz' else
                    ('mgh','.mgh') if output_format == 'mgh' else
                    ('freesurfer_morph',''))
    # make the filemap...
    worklog('Preparing Filemap...')
    fm = {}
    for (h,cl) in six.iteritems(clean_retinotopy if not no_clean else {}):
        for (k,val) in six.iteritems(cl):
            flnm = ('%s.clean_%s.' + output_format) % (h,k)
            fm[flnm] = val
    for (h,cm) in six.iteritems(radtan_cmag if radtan_cmag is not None else {}):
        flnm = ('%s.rad_cmag.' % h) + output_format
        fm[flnm] = cm[0]
        flnm = ('%s.tan_cmag.' % h) + output_format
        fm[flnm] = cm[1]
    for (h,cm) in six.iteritems(spotlight_cmag if spotlight_cmag is not None else {}):
        flnm = ('%s.spot_cmag.' % h) + output_format
        fm[flnm] = cm
    # okay, make that a persistent map:
    filemap = pimms.persist(fm)
    output_path = os.path.expanduser(os.path.expandvars(output_path))
    # the function for exporting all properties:
    def export_all():
        '''
        This function will export all files from its associated filemap and return a list of the
        filenames.
        '''
        if not os.path.isdir(output_path):
            if not create_directory:
                raise ValueError('No such path and create_direcotry is False: %s' % output_path)
            os.makedirs(os.path.abspath(output_path), 0o755)
        filenames = []
        worklog('Writing Files...')
        wl = worklog.indent()
        for (flnm,val) in six.iteritems(filemap):
            flnm = os.path.join(output_path, flnm)
            wl(flnm)
            filenames.append(nyio.save(flnm, val, fmt))
        return filenames
    return {'filemap': filemap, 'export_all_fn': export_all}

retinotopy_plan_data = pyr.pmap(
    {'init_worklog':calc_worklog,
     'init_subject':calc_subject,
     'init_hemispheres':calc_hemispheres,
     'init_labels':calc_labels,
     'init_retinotopy':calc_retinotopy,
     'cleaning':calc_clean_maps,
     'spotlight_cmag':calc_spotlight_cmag,
     'radtan_cmag':calc_radtan_cmag,
     'filemap':calc_filemap})
retinotopy_plan = pimms.plan(retinotopy_plan_data)

retinotopy_cmdline_abbrevs = {'output_format':    'f',
                              'overwrite':        'o',
                              'create_directory': 'c',
                              'hemispheres':      'H',
                              'labels':           'l',
                              'angles':           'a',
                              'eccens':           'e',
                              'weights':          'w',
                              'output_path':      'o',
                              'verbose':          'v'}

def _format_afferent_doc(docstr, abbrevs=None, cols=80):
    try:
        (ln1, docs) = docstr.split('\n\n')
    except Exception: return ''
    anm0 = ln1.split(' (')[0]
    anm = anm0.replace('_', '-')
    header = '  --' + anm
    if abbrevs and anm0 in abbrevs: header = header + ' | -' + abbrevs[anm0]
    docs = [
        '\n     '.join(
            textwrap.wrap(
                ' '.join([s.strip() for s in ss.split('\n')[1:]]),
                cols-6))
        for ss in docs.split('\n)')]
    return header + '\n' + ''.join(['   * ' + d for d in docs])

info = \
    '''SYNTAX:  python -m neuropythy retinotopy <subject-id>

Neuropythy's retinotopy command performs cleaning of a set of retinotopic
maps for a subject then calculates cortical magnification.

The following optional arguments may be provided:
  --help | -h
    * Prints this help message.
''' + '\n'.join(
    [s for s in [_format_afferent_doc(retinotopy_plan.afferent_docs[af],
                                      retinotopy_cmdline_abbrevs)
                 for af in retinotopy_plan.afferents
                 if af in retinotopy_plan.afferent_docs]
     if len(s) > 0])

def main(*argv):
    '''
    neuropythy.commands.retinotopy.main() runs the main function of the retinotopy command in the
       neuropythy library. See the `python -m neuropythy retinotopy --help` or atlas.info for more
       information.
    '''
    argv = [aa for arg in argv for aa in (arg if pimms.is_vector(arg) else [arg])]
    imap = pimms.argv_parse(
        retinotopy_plan, argv,
        arg_abbrevs=retinotopy_cmdline_abbrevs)
    argv = imap['argv']
    if len(argv) == 0 or '--help' in argv or '-h' in argv:
        print(info)
        return 1
    try: imap['export_all_fn']()
    except Exception as e:
        raise
        sys.stderr.write('\nERROR:\n' + str(e) + '\n')
        sys.stderr.flush()
        sys.exit(2)
    return 0



####################################################################################################
# neuropythy/datasets/__init__.py
# Datasets for neuropythy.
# by Noah C. Benson

# mainly just to force these to load when datasets is loaded:
from .benson_winawer_2018       import (BensonWinawer2018Dataset)
from .hcp                       import (HCPDataset, HCPRetinotopyDataset, HCPMetaDataset)
from .visual_performance_fields import (VisualPerformanceFieldsDataset)
from .hcp_lines                 import (HCPLinesDataset)
# TODO: https://openneuro.org/crn/datasets/ds001499/snapshots/1.1.0/download -- add the BOLD5000
#     : dataset to neuropythy (see bold5000.org)
# import this last so that we get the most updated version of data
from .core import (data, Dataset)





#! /usr/bin/env python

import os
from setuptools import (setup, Extension)

# Deduce the version from the __init__.py file:
version = None
with open(os.path.join(os.path.dirname(__file__), 'neuropythy', '__init__.py'), 'r') as fid:
    for line in (line.strip() for line in fid):
        if line.startswith('__version__'):
            version = line.split('=')[1].strip().strip('\'')
            break
if version is None: raise ValueError('No version found in neuropythy/__init__.py!')

setup(
    name='neuropythy',
    version=version,
    description='Toolbox for flexible cortical mesh analysis and registration',
    keywords='neuroscience mesh cortex registration',
    author='Noah C. Benson',
    author_email='nben@uw.edu',
    maintainer_email='nben@uw.edu',
    long_description='''
                     See the README.md file at the github repository for this package:
                     https://github.com/noahbenson/neuropythy
                     ''',
    url='https://github.com/noahbenson/neuropythy',
    download_url='https://github.com/noahbenson/neuropythy',
    license='AGPL',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        # Removing support for Python 2, since it's well past EOL.
        #'Programming Language :: Python :: 2',
        #'Programming Language :: Python :: 2.7', 
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS'],
    packages=['neuropythy',
              'neuropythy.util',
              'neuropythy.math',
              'neuropythy.java',
              'neuropythy.io',
              'neuropythy.geometry',
              'neuropythy.optimize',
              'neuropythy.mri',
              'neuropythy.freesurfer',
              'neuropythy.hcp',
              'neuropythy.registration',
              'neuropythy.vision',
              'neuropythy.graphics',
              'neuropythy.datasets',
              'neuropythy.plans',
              'neuropythy.commands',
              'neuropythy.test'],
    # not part of library; just included as an example of how this would work
    #ext_modules=[Extension('neuropythy.c_label', sources=['src/c_label.c'],
    #                       include_dirs=[np.get_include()])],
    include_package_data=True,
    package_data={
        '': ['LICENSE.txt',
             'neuropythy/lib/nben/target/nben-standalone.jar',
             'neuropythy/lib/models/v123.fmm.gz',
             'neuropythy/lib/models/lh.benson17.fmm.gz',
             'neuropythy/lib/models/rh.benson17.fmm.gz',
             'neuropythy/lib/projections/lh.occipital_pole.mp.json',
             'neuropythy/lib/projections/rh.occipital_pole.mp.json',
             'neuropythy/lib/data/fsaverage/surf/lh.benson14_angle.v4_0.mgz',
             'neuropythy/lib/data/fsaverage/surf/lh.benson14_sigma.v4_0.mgz',
             'neuropythy/lib/data/fsaverage/surf/lh.benson14_varea.v4_0.mgz',
             'neuropythy/lib/data/fsaverage/surf/lh.benson14_eccen.v4_0.mgz',
             'neuropythy/lib/data/fsaverage/surf/rh.benson14_retinotopy.v4_0.sphere.reg',
             'neuropythy/lib/data/fsaverage/surf/lh.wang15_mplbl.v1_0.mgz',
             'neuropythy/lib/data/fsaverage/surf/rh.benson14_varea.v4_0.mgz',
             'neuropythy/lib/data/fsaverage/surf/rh.benson14_eccen.v4_0.mgz',
             'neuropythy/lib/data/fsaverage/surf/rh.wang15_mplbl.v1_0.mgz',
             'neuropythy/lib/data/fsaverage/surf/rh.benson14_angle.v4_0.mgz',
             'neuropythy/lib/data/fsaverage/surf/rh.benson14_sigma.v4_0.mgz',
             'neuropythy/lib/data/fsaverage/surf/lh.benson14_retinotopy.v4_0.sphere.reg',
             'neuropythy/lib/data/fsaverage/surf/lh.rosenke18_vcatlas.v1_0.mgz',
             'neuropythy/lib/data/fsaverage/surf/rh.rosenke18_vcatlas.v1_0.mgz',
             'neuropythy/lib/data/fsaverage/surf/lh.glasser16_atlas.v1_0.mgz',
             'neuropythy/lib/data/fsaverage/surf/rh.glasser16_atlas.v1_0.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_angle.v2_0.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_angle.v2_1.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_eccen.v3_0.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_enorm.v1_0.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_angle.v2_5.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_varea.v3_0.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_eccen.v1_0.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_angle.v3_0.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_eccen.v2_5.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_sigma.v3_0.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_varea.v2_5.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_angle.v1_0.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_varea.v2_0.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_varea.v2_1.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_eccen.v2_0.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_eccen.v2_1.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_varea.v1.0.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_anorm.v1_0.mgz',
             'neuropythy/lib/data/fsaverage_sym/surf/lh.benson14_retinotopy.v3_0.sphere.reg',
             'neuropythy/lib/data/fs_LR/lh.atlasroi.32k_fs_LR.shape.gii',
             'neuropythy/lib/data/fs_LR/rh.atlasroi.32k_fs_LR.shape.gii',
             'neuropythy/lib/data/fs_LR/lh.atlasroi.59k_fs_LR.shape.gii',
             'neuropythy/lib/data/fs_LR/rh.atlasroi.59k_fs_LR.shape.gii',
             'neuropythy/lib/data/fs_LR/lh.atlasroi.164k_fs_LR.shape.gii',
             'neuropythy/lib/data/fs_LR/rh.atlasroi.164k_fs_LR.shape.gii',
             'neuropythy/lib/data/hcp_lines_osftree.json.gz']},
    install_requires=['numpy>=1.13',
                      'scipy>=1.1',
                      'six >= 1.13',
                      'nibabel>=2.0',
                      'pyrsistent>=0.11',
                      'pint>=0.7',
                      'pimms>=0.3.16',
                      'py4j>=0.10',
                      'h5py>=2.8.0',
                      's3fs>=0.1.5'],
    extras_require={
        'graphics2D': ['matplotlib>=1.5.3'],
        'graphics3D': ['matplotlib>=1.5.3', 'ipyvolume>=0.5.1'],
        'torch':      ['torch>=1.6.0'],
        'all':        ['matplotlib>=1.5.3', 'ipyvolume>=0.5.1', 'torch>=1.6.0']})



####################################################################################################
# __main__.py
# The main function, if neuropythy is invoked directly as command.
# By Noah C. Benson

import sys

from neuropythy.commands import (commands)

def main(argv):
    if len(argv) < 1:
        return 0
    if argv[0] not in commands:
        sys.stderr.write('The given command \'' + argv[0] + '\' not recognized.\n')
        return 1
    return commands[argv[0]](argv[1:])

# Run the main function
sys.exit(main(sys.argv[1:]))
